@@ -41,16 +41,63 @@ isinteractive() = (is_interactive::Bool)
 
 ## package depots (registries, packages, environments) ##
 
+"""
+    DEPOT_PATH
+
+A stack of "depot" locations where the package manager, as well as Julia's code
+loading mechanisms, look for package registries, installed packages, named
+environments, repo clones, cached compiled package images, and configuration
+files. By default it includes:
+
+1. `~/.julia` where `~` is the user home as appropriate on the system;
+2. an architecture-specific shared system directory, e.g. `/usr/local/share/julia`;
+3. an architecture-independent shared system directory, e.g. `/usr/share/julia`.
+
+So `DEPOT_PATH` might be:
+```julia
+[joinpath(homedir(), ".julia"), "/usr/local/share/julia", "/usr/share/julia"]
+```
+The first entry is the "user depot" and should be writable by and owned by the
+current user. The user depot is where: registries are cloned, new package versions
+are installed, named environments are created and updated, package repos are cloned,
+newly compiled package image files are saved, log files are written, development
+packages are checked out by default, and global configuration data is saved. Later
+entries in the depot path are treated as read-only and are appropriate for
+registries, packages, etc. installed and managed by system administrators.
+
+`DEPOT_PATH` is populated based on the [`JULIA_DEPOT_PATH`](@ref JULIA_DEPOT_PATH)
+environment variable if set.
+
+See also:
+[`JULIA_DEPOT_PATH`](@ref JULIA_DEPOT_PATH), and
+[Code Loading](@ref Code-Loading).
+"""
 const DEPOT_PATH = String[]
 
+function append_default_depot_path!(DEPOT_PATH)
+    path = joinpath(homedir(), ".julia")
+    path in DEPOT_PATH || push!(DEPOT_PATH, path)
+    path = abspath(Sys.BINDIR, "..", "local", "share", "julia")
+    path in DEPOT_PATH || push!(DEPOT_PATH, path)
+    path = abspath(Sys.BINDIR, "..", "share", "julia")
+    path in DEPOT_PATH || push!(DEPOT_PATH, path)
+end
+
 function init_depot_path()
+    empty!(DEPOT_PATH)
     if haskey(ENV, "JULIA_DEPOT_PATH")
-        depots = split(ENV["JULIA_DEPOT_PATH"], Sys.iswindows() ? ';' : ':')
-        append!(empty!(DEPOT_PATH), map(expanduser, depots))
+        str = ENV["JULIA_DEPOT_PATH"]
+        isempty(str) && return
+        for path in split(str, Sys.iswindows() ? ';' : ':')
+            if isempty(path)
+                append_default_depot_path!(DEPOT_PATH)
+            else
+                path = expanduser(path)
+                path in DEPOT_PATH || push!(DEPOT_PATH, path)
+            end
+        end
     else
-        push!(empty!(DEPOT_PATH), joinpath(homedir(), ".julia"))
-        push!(DEPOT_PATH, abspath(Sys.BINDIR, "..", "local", "share", "julia"))
-        push!(DEPOT_PATH, abspath(Sys.BINDIR, "..", "share", "julia"))
+        append_default_depot_path!(DEPOT_PATH)
     end
 end
 
@@ -76,7 +123,36 @@ const DEFAULT_LOAD_PATH = ["@", "@v#.#", "@stdlib"]
     LOAD_PATH
 
 An array of paths for `using` and `import` statements to consider as project
-environments or package directories when loading code. See [Code Loading](@ref Code-Loading).
+environments or package directories when loading code. It is populated based on
+the [`JULIA_LOAD_PATH`](@ref JULIA_LOAD_PATH) environment variable if set;
+otherwise it defaults to `["@", "@v#.#", "@stdlib"]`. Entries starting with `@`
+have special meanings:
+
+- `@` refers to the "current active environment", the initial value of which is
+  initially determined by the [`JULIA_PROJECT`](@ref JULIA_PROJECT) environment
+  variable or the `--project` command-line option.
+
+- `@stdlib` expands to the absolute path of the current Julia installation's
+  standard library directory.
+
+- `@name` refers to a named environment, which are stored in depots (see
+  [`JULIA_DEPOT_PATH`](@ref JULIA_DEPOT_PATH)) under the `environments`
+  subdirectory. The user's named environments are stored in
+  `~/.julia/environments` so `@name` would refer to the environment in
+  `~/.julia/environments/name` if it exists and contains a `Project.toml` file.
+  If `name` contains `#` characters, then they are replaced with the major, minor
+  and patch components of the Julia version number. For example, if you are
+  running Julia 1.2 then `@v#.#` expands to `@v1.2` and will look for an
+  environment by that name, typically at `~/.julia/environments/v1.2`.
+
+The fully expanded value of `LOAD_PATH` that is searched for projects and packages
+can be seen by calling the `Base.load_path()` function.
+
+See also:
+[`JULIA_LOAD_PATH`](@ref JULIA_LOAD_PATH),
+[`JULIA_PROJECT`](@ref JULIA_PROJECT),
+[`JULIA_DEPOT_PATH`](@ref JULIA_DEPOT_PATH), and
+[Code Loading](@ref Code-Loading).
 """
 const LOAD_PATH = copy(DEFAULT_LOAD_PATH)
 const HOME_PROJECT = Ref{Union{String,Nothing}}(nothing)
@@ -119,6 +195,7 @@ function parse_load_path(str::String)
                 env = current_project()
                 env === nothing && continue
             end
+            env = expanduser(env)
             env in envs || push!(envs, env)
         end
     end
@@ -138,7 +215,7 @@ function init_load_path()
         unsafe_string(Base.JLOptions().project) :
         get(ENV, "JULIA_PROJECT", nothing))
     HOME_PROJECT[] =
-        project == nothing ? nothing :
+        project === nothing ? nothing :
         project == "" ? nothing :
         project == "@." ? current_project() : abspath(project)
     append!(empty!(LOAD_PATH), paths)
