@@ -47,6 +47,7 @@ JL_DLLEXPORT jl_methtable_t *jl_new_method_table(jl_sym_t *name, jl_module_t *mo
     mt->kwsorter = NULL;
     mt->backedges = NULL;
     JL_MUTEX_INIT(&mt->writelock);
+    mt->offs = 1;
     return mt;
 }
 
@@ -480,6 +481,8 @@ JL_DLLEXPORT jl_datatype_t *jl_new_datatype(
             if (!abstract) {
                 tn->mt = jl_new_method_table(name, module);
                 jl_gc_wb(tn, tn->mt);
+                if (jl_svec_len(parameters) > 0)
+                    tn->mt->offs = 0;
             }
         }
         t->name = tn;
@@ -797,8 +800,24 @@ JL_DLLEXPORT jl_value_t *jl_new_struct(jl_datatype_t *type, ...)
     return jv;
 }
 
-JL_DLLEXPORT jl_value_t *jl_new_structv(jl_datatype_t *type, jl_value_t **args,
-                                        uint32_t na)
+static void init_struct_tail(jl_datatype_t *type, jl_value_t *jv, size_t na)
+{
+    size_t nf = jl_datatype_nfields(type);
+    for(size_t i=na; i < nf; i++) {
+        if (jl_field_isptr(type, i)) {
+            *(jl_value_t**)((char*)jl_data_ptr(jv)+jl_field_offset(type,i)) = NULL;
+        }
+        else {
+            jl_value_t *ft = jl_field_type(type, i);
+            if (jl_is_uniontype(ft)) {
+                uint8_t *psel = &((uint8_t *)jv)[jl_field_offset(type, i) + jl_field_size(type, i) - 1];
+                *psel = 0;
+            }
+        }
+    }
+}
+
+JL_DLLEXPORT jl_value_t *jl_new_structv(jl_datatype_t *type, jl_value_t **args, uint32_t na)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     if (type->instance != NULL) {
@@ -811,7 +830,6 @@ JL_DLLEXPORT jl_value_t *jl_new_structv(jl_datatype_t *type, jl_value_t **args,
     }
     if (type->layout == NULL)
         jl_type_error("new", (jl_value_t*)jl_datatype_type, (jl_value_t*)type);
-    size_t nf = jl_datatype_nfields(type);
     jl_value_t *jv = jl_gc_alloc(ptls, jl_datatype_size(type), type);
     JL_GC_PUSH1(&jv);
     for (size_t i = 0; i < na; i++) {
@@ -820,17 +838,41 @@ JL_DLLEXPORT jl_value_t *jl_new_structv(jl_datatype_t *type, jl_value_t **args,
             jl_type_error("new", ft, args[i]);
         jl_set_nth_field(jv, i, args[i]);
     }
-    for(size_t i=na; i < nf; i++) {
-        if (jl_field_isptr(type, i)) {
-            *(jl_value_t**)((char*)jl_data_ptr(jv)+jl_field_offset(type,i)) = NULL;
-        }
-        else {
+    init_struct_tail(type, jv, na);
+    JL_GC_POP();
+    return jv;
+}
+
+JL_DLLEXPORT jl_value_t *jl_new_structt(jl_datatype_t *type, jl_value_t *tup)
+{
+    jl_ptls_t ptls = jl_get_ptls_states();
+    if (!jl_is_tuple(tup))
+        jl_type_error("new", (jl_value_t*)jl_tuple_type, tup);
+    size_t nargs = jl_nfields(tup);
+    size_t nf = jl_datatype_nfields(type);
+    JL_NARGS(new, nf, nf);
+    if (type->instance != NULL) {
+        jl_datatype_t *tupt = (jl_datatype_t*)jl_typeof(tup);
+        for (size_t i = 0; i < nargs; i++) {
             jl_value_t *ft = jl_field_type(type, i);
-            if (jl_is_uniontype(ft)) {
-                uint8_t *psel = &((uint8_t *)jv)[jl_field_offset(type, i) + jl_field_size(type, i) - 1];
-                *psel = 0;
-            }
+            jl_value_t *et = jl_field_type(tupt, i);
+            assert(jl_is_concrete_type(ft) && jl_is_concrete_type(et));
+            if (et != ft)
+                jl_type_error("new", ft, jl_get_nth_field(tup, i));
         }
+        return type->instance;
+    }
+    if (type->layout == NULL)
+        jl_type_error("new", (jl_value_t*)jl_datatype_type, (jl_value_t*)type);
+    jl_value_t *jv = jl_gc_alloc(ptls, jl_datatype_size(type), type);
+    jl_value_t *fi = NULL;
+    JL_GC_PUSH2(&jv, &fi);
+    for (size_t i = 0; i < nargs; i++) {
+        jl_value_t *ft = jl_field_type(type, i);
+        fi = jl_get_nth_field(tup, i);
+        if (!jl_isa(fi, ft))
+            jl_type_error("new", ft, fi);
+        jl_set_nth_field(jv, i, fi);
     }
     JL_GC_POP();
     return jv;
