@@ -3,7 +3,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <assert.h>
 #include <sys/stat.h>
 
 #include "platform.h"
@@ -16,6 +15,7 @@
 #include <unistd.h>
 #include <dlfcn.h>
 #endif
+#include "julia_assert.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -56,22 +56,22 @@ static int endswith_extension(const char *path)
 
 #define PATHBUF 512
 
-extern char *julia_home;
+extern char *julia_bindir;
 
 #define JL_RTLD(flags, FLAG) (flags & JL_RTLD_ ## FLAG ? RTLD_ ## FLAG : 0)
 
-static void JL_NORETURN jl_dlerror(const char *fmt, const char *sym)
+static const char * jl_dlerror(void)
 {
 #ifdef _OS_WINDOWS_
-    CHAR reason[256];
+    static JL_THREAD_LOCAL CHAR reason[256];
     FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
             NULL, GetLastError(),
             MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
             reason, sizeof(reason) / sizeof(reason[0]), NULL);
+    return (const char *)&reason[0];
 #else
-    const char *reason = dlerror();
+    return dlerror();
 #endif
-    jl_errorf(fmt, sym, reason);
 }
 
 JL_DLLEXPORT void *jl_dlopen(const char *filename, unsigned flags)
@@ -109,7 +109,7 @@ JL_DLLEXPORT int jl_dlclose(void *handle)
 {
 #ifdef _OS_WINDOWS_
     if (!handle) return -1;
-    return FreeLibrary((HMODULE) handle);
+    return !FreeLibrary((HMODULE) handle);
 #else
     dlerror(); /* Reset error status. */
     if (!handle) return -1;
@@ -117,7 +117,7 @@ JL_DLLEXPORT int jl_dlclose(void *handle)
 #endif
 }
 
-static void *jl_load_dynamic_library_(const char *modname, unsigned flags, int throw_err)
+JL_DLLEXPORT void *jl_load_dynamic_library(const char *modname, unsigned flags, int throw_err)
 {
     char path[PATHBUF];
     int i;
@@ -134,13 +134,13 @@ static void *jl_load_dynamic_library_(const char *modname, unsigned flags, int t
     if (modname == NULL) {
 #ifdef _OS_WINDOWS_
         if (!GetModuleHandleExW(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
-                                (LPCWSTR)(&jl_load_dynamic_library),
+                                (LPCWSTR)(uintptr_t)(&jl_load_dynamic_library),
                                 (HMODULE*)&handle)) {
             jl_error("could not load base module");
         }
 #else
         Dl_info info;
-        if (!dladdr(&jl_load_dynamic_library, &info) || !info.dli_fname)
+        if (!dladdr((void*)(uintptr_t)&jl_load_dynamic_library, &info) || !info.dli_fname)
             jl_error("could not load base module");
         handle = dlopen(info.dli_fname, RTLD_NOW);
 #endif
@@ -191,66 +191,65 @@ static void *jl_load_dynamic_library_(const char *modname, unsigned flags, int t
             goto done;
     }
 
-#if defined(__linux__) || defined(__FreeBSD__)
-    // check map of versioned libs from "libX" to full soname "libX.so.ver"
-    if (!abspath && n_extensions > 1) { // soname map only works for libX
-        handle = jl_dlopen_soname(modname, strlen(modname), flags);
-        if (handle)
-            goto done;
-    }
-#endif
-
 notfound:
-    if (throw_err)
-        jl_dlerror("could not load library \"%s\"\n%s", modname);
+    if (throw_err) {
+        const char * reason = jl_dlerror();
+        jl_errorf("could not load library \"%s\"\n%s", modname, reason);
+    }
     return NULL;
 
 done:
     return handle;
 }
 
-JL_DLLEXPORT void *jl_load_dynamic_library_e(const char *modname, unsigned flags)
+JL_DLLEXPORT int jl_dlsym(void *handle, const char *symbol, void ** value, int throw_err)
 {
-    return jl_load_dynamic_library_(modname, flags, 0);
-}
+    int symbol_found = 0;
 
-JL_DLLEXPORT void *jl_load_dynamic_library(const char *modname, unsigned flags)
-{
-    return jl_load_dynamic_library_(modname, flags, 1);
-}
-
-JL_DLLEXPORT void *jl_dlsym_e(void *handle, const char *symbol)
-{
+    /* First, get the symbol value */
 #ifdef _OS_WINDOWS_
-    void *ptr = GetProcAddress((HMODULE) handle, symbol);
+    *value = GetProcAddress((HMODULE) handle, symbol);
 #else
     dlerror(); /* Reset error status. */
-    void *ptr = dlsym(handle, symbol);
+    *value = dlsym(handle, symbol);
 #endif
-    return ptr;
-}
 
-JL_DLLEXPORT void *jl_dlsym(void *handle, const char *symbol)
-{
-    void *ptr = jl_dlsym_e(handle, symbol);
-    if (!ptr)
-        jl_dlerror("could not load symbol \"%s\":\n%s", symbol);
-    return ptr;
+    /* Next, check for errors.  On Windows, a NULL pointer means the symbol
+     * was not found.  On everything else, we can have NULL symbols, so we check
+     * for non-NULL returns from dlerror().  Note that we unconditionally call
+     * jl_dlerror() on POSIX systems, but on Windows systems we only call it
+     * when we have been returned a NULL symbol.*/
+    const char * err = NULL;
+#ifdef _OS_WINDOWS_
+    symbol_found = *value != NULL;
+#else
+    err = jl_dlerror();
+    symbol_found = err == NULL;
+#endif
+
+    if (!symbol_found && throw_err) {
+#ifdef _OS_WINDOWS_
+        err = jl_dlerror();
+#endif
+        jl_errorf("could not load symbol \"%s\":\n%s", symbol, err);
+    }
+    return symbol_found;
 }
 
 #ifdef _OS_WINDOWS_
 //Look for symbols in win32 libraries
 const char *jl_dlfind_win32(const char *f_name)
 {
-    if (jl_dlsym_e(jl_exe_handle, f_name))
+    void * dummy;
+    if (jl_dlsym(jl_exe_handle, f_name, &dummy, 0))
         return JL_EXE_LIBNAME;
-    if (jl_dlsym_e(jl_dl_handle, f_name))
+    if (jl_dlsym(jl_dl_handle, f_name, &dummy, 0))
         return JL_DL_LIBNAME;
-    if (jl_dlsym_e(jl_kernel32_handle, f_name))
+    if (jl_dlsym(jl_kernel32_handle, f_name, &dummy, 0))
         return "kernel32";
-    if (jl_dlsym_e(jl_ntdll_handle, f_name))
+    if (jl_dlsym(jl_ntdll_handle, f_name, &dummy, 0))
         return "ntdll";
-    if (jl_dlsym_e(jl_crtdll_handle, f_name))
+    if (jl_dlsym(jl_crtdll_handle, f_name, &dummy, 0))
 #if defined(_MSC_VER)
 #if _MSC_VER == 1800
         return "msvcr120";
@@ -260,7 +259,7 @@ const char *jl_dlfind_win32(const char *f_name)
 #else
         return "msvcrt";
 #endif
-    if (jl_dlsym_e(jl_winsock_handle, f_name))
+    if (jl_dlsym(jl_winsock_handle, f_name, &dummy, 0))
         return "ws2_32";
     // additional common libraries (libc?) could be added here, but in general,
     // it is better to specify the library explicitly in the code. This exists
