@@ -7,7 +7,6 @@
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
-#include <assert.h>
 
 #ifdef _OS_WINDOWS_
 #include <ws2tcpip.h>
@@ -28,6 +27,8 @@
 #define write _write
 #endif
 
+#include "julia_assert.h"
+
 #ifdef __cplusplus
 #include <cstring>
 extern "C" {
@@ -35,16 +36,6 @@ extern "C" {
 
 static uv_async_t signal_async;
 
-#ifdef _OS_WINDOWS_
-// uv_async_t is buggy on windows. Initializing one breaks the sysimg build.
-void jl_wake_libuv(void)
-{
-}
-
-void jl_init_signal_async(void)
-{
-}
-#else
 static void jl_signal_async_cb(uv_async_t *hdl)
 {
     // This should abort the current loop and the julia code it returns to
@@ -58,11 +49,27 @@ void jl_wake_libuv(void)
     uv_async_send(&signal_async);
 }
 
-void jl_init_signal_async(void)
+jl_mutex_t jl_uv_mutex;
+
+void jl_init_uv(void)
 {
     uv_async_init(jl_io_loop, &signal_async, jl_signal_async_cb);
+    JL_MUTEX_INIT(&jl_uv_mutex); // a file-scope initializer can be used instead
+}
+
+int jl_uv_n_waiters = 0;
+
+void JL_UV_LOCK(void)
+{
+    if (jl_mutex_trylock(&jl_uv_mutex)) {
+    }
+    else {
+        jl_atomic_fetch_add(&jl_uv_n_waiters, 1);
+        jl_wake_libuv();
+        JL_LOCK(&jl_uv_mutex);
+        jl_atomic_fetch_add(&jl_uv_n_waiters, -1);
+    }
 }
-#endif
 
 void jl_uv_call_close_callback(jl_value_t *val)
 {
@@ -88,6 +95,7 @@ static void jl_uv_closeHandle(uv_handle_t *handle)
     // also let the client app do its own cleanup
     if (handle->type != UV_FILE && handle->data) {
         size_t last_age = jl_get_ptls_states()->world_age;
+        // TODO: data race on jl_world_counter across many files, to be fixed in a separate revision
         jl_get_ptls_states()->world_age = jl_world_counter;
         jl_uv_call_close_callback((jl_value_t*)handle->data);
         jl_get_ptls_states()->world_age = last_age;
@@ -97,80 +105,123 @@ static void jl_uv_closeHandle(uv_handle_t *handle)
     free(handle);
 }
 
-static void jl_uv_shutdownCallback(uv_shutdown_t *req, int status)
+static void jl_uv_flush_close_callback(uv_write_t *req, int status)
 {
-    /*
-     * This happens if the remote machine closes the connecition while we're
-     * in the shutdown request (in that case we call uv_close, thus cancelling
-     * this request).
-     */
-    if (status != UV__ECANCELED && !uv_is_closing((uv_handle_t*)req->handle)) {
-        uv_close((uv_handle_t*)req->handle, &jl_uv_closeHandle);
+    uv_stream_t *stream = req->handle;
+    req->handle = NULL;
+    // ignore attempts to close the stream while attempting a graceful shutdown
+#ifdef _OS_WINDOWS_
+    if (stream->stream.conn.shutdown_req)
+#else
+    if (stream->shutdown_req)
+#endif
+    {
+        free(req);
+        return;
+    }
+    if (status == 0 && uv_is_writable(stream) && stream->write_queue_size != 0) {
+        // new data was written, wait for it to flush too
+        uv_buf_t buf;
+        buf.base = (char*)(req + 1);
+        buf.len = 0;
+        req->data = NULL;
+        if (uv_write(req, stream, &buf, 1, (uv_write_cb)jl_uv_flush_close_callback) == 0)
+            return;
+    }
+    if (!uv_is_closing((uv_handle_t*)stream)) { // avoid double-close on the stream
+        if (stream->type == UV_TTY)
+            uv_tty_set_mode((uv_tty_t*)stream, UV_TTY_MODE_NORMAL);
+        uv_close((uv_handle_t*)stream, &jl_uv_closeHandle);
     }
     free(req);
 }
 
+static void uv_flush_callback(uv_write_t *req, int status)
+{
+    *(int*)(req->data) = 1;
+    uv_stop(req->handle->loop);
+    free(req);
+}
+
+// Turn a normal write into a blocking write (primarily for use from C and gdb).
+// Warning: This calls uv_run, so it can have unbounded side-effects.
+// Be care where you call it from! - the libuv loop is also not reentrant.
+void jl_uv_flush(uv_stream_t *stream)
+{
+    if (stream == (void*)STDIN_FILENO ||
+        stream == (void*)STDOUT_FILENO ||
+        stream == (void*)STDERR_FILENO)
+        return;
+    if (stream->type != UV_TTY &&
+        stream->type != UV_TCP &&
+        stream->type != UV_NAMED_PIPE)
+        return;
+    JL_UV_LOCK();
+    while (uv_is_writable(stream) && stream->write_queue_size != 0) {
+        int fired = 0;
+        uv_buf_t buf;
+        buf.base = (char*)(&buf + 1);
+        buf.len = 0;
+        uv_write_t *write_req = (uv_write_t*)malloc(sizeof(uv_write_t));
+        write_req->data = (void*)&fired;
+        if (uv_write(write_req, stream, &buf, 1, uv_flush_callback) != 0) {
+            JL_UV_UNLOCK();
+            return;
+        }
+        while (!fired) {
+            uv_run(uv_default_loop(), UV_RUN_DEFAULT);
+        }
+    }
+    JL_UV_UNLOCK();
+}
+
 // getters and setters
+// TODO: check if whoever calls these is thread-safe
+JL_DLLEXPORT int jl_uv_process_pid(uv_process_t *p) { return p->pid; }
 JL_DLLEXPORT void *jl_uv_process_data(uv_process_t *p) { return p->data; }
 JL_DLLEXPORT void *jl_uv_buf_base(const uv_buf_t *buf) { return buf->base; }
 JL_DLLEXPORT size_t jl_uv_buf_len(const uv_buf_t *buf) { return buf->len; }
 JL_DLLEXPORT void jl_uv_buf_set_base(uv_buf_t *buf, char *b) { buf->base = b; }
 JL_DLLEXPORT void jl_uv_buf_set_len(uv_buf_t *buf, size_t n) { buf->len = n; }
 JL_DLLEXPORT void *jl_uv_connect_handle(uv_connect_t *connect) { return connect->handle; }
-JL_DLLEXPORT void *jl_uv_getaddrinfo_data(uv_getaddrinfo_t *req) { return req->data; }
-JL_DLLEXPORT uv_file jl_uv_file_handle(jl_uv_file_t *f) { return f->file; }
+JL_DLLEXPORT uv_os_fd_t jl_uv_file_handle(jl_uv_file_t *f) { return f->file; }
 JL_DLLEXPORT void *jl_uv_req_data(uv_req_t *req) { return req->data; }
 JL_DLLEXPORT void jl_uv_req_set_data(uv_req_t *req, void *data) { req->data = data; }
 JL_DLLEXPORT void *jl_uv_handle_data(uv_handle_t *handle) { return handle->data; }
 JL_DLLEXPORT void *jl_uv_write_handle(uv_write_t *req) { return req->handle; }
 
+extern volatile unsigned _threadedregion;
+
 JL_DLLEXPORT int jl_run_once(uv_loop_t *loop)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     if (loop) {
-        loop->stop_flag = 0;
         jl_gc_safepoint_(ptls);
-        return uv_run(loop,UV_RUN_ONCE);
-    }
-    else return 0;
-}
-
-JL_DLLEXPORT void jl_run_event_loop(uv_loop_t *loop)
-{
-    jl_ptls_t ptls = jl_get_ptls_states();
-    if (loop) {
+        JL_UV_LOCK();
         loop->stop_flag = 0;
-        jl_gc_safepoint_(ptls);
-        uv_run(loop,UV_RUN_DEFAULT);
+        int r = uv_run(loop,UV_RUN_ONCE);
+        JL_UV_UNLOCK();
+        return r;
     }
+    return 0;
 }
 
 JL_DLLEXPORT int jl_process_events(uv_loop_t *loop)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
-    if (loop) {
-        loop->stop_flag = 0;
+    if (loop && (_threadedregion || ptls->tid == 0)) {
         jl_gc_safepoint_(ptls);
-        return uv_run(loop,UV_RUN_NOWAIT);
+        if (jl_mutex_trylock(&jl_uv_mutex)) {
+            loop->stop_flag = 0;
+            int r = uv_run(loop,UV_RUN_NOWAIT);
+            JL_UV_UNLOCK();
+            return r;
+        }
     }
-    else return 0;
+    return 0;
 }
 
-JL_DLLEXPORT int jl_init_pipe(uv_pipe_t *pipe, int writable, int readable,
-                              int julia_only)
-{
-     int flags = 0;
-     if (writable)
-         flags |= UV_PIPE_WRITABLE;
-     if (readable)
-         flags |= UV_PIPE_READABLE;
-     if (!julia_only)
-         flags |= UV_PIPE_SPAWN_SAFE;
-     int err = uv_pipe_init(jl_io_loop, pipe, flags);
-     return err;
-}
-
-static void jl_proc_exit_cleanup(uv_process_t *process, int64_t exit_status, int term_signal)
+static void jl_proc_exit_cleanup_cb(uv_process_t *process, int64_t exit_status, int term_signal)
 {
     uv_close((uv_handle_t*)process, (uv_close_cb)&free);
 }
@@ -181,59 +232,45 @@ JL_DLLEXPORT void jl_close_uv(uv_handle_t *handle)
         // take ownership of this handle,
         // so we can waitpid for the resource to exit and avoid leaving zombies
         assert(handle->data == NULL); // make sure Julia has forgotten about it already
-        ((uv_process_t*)handle)->exit_cb = jl_proc_exit_cleanup;
+        ((uv_process_t*)handle)->exit_cb = jl_proc_exit_cleanup_cb;
         return;
     }
-
+    JL_UV_LOCK();
     if (handle->type == UV_FILE) {
         uv_fs_t req;
         jl_uv_file_t *fd = (jl_uv_file_t*)handle;
-        if (fd->file != -1) {
+        if ((ssize_t)fd->file != -1) {
             uv_fs_close(handle->loop, &req, fd->file, NULL);
-            fd->file = -1;
+            fd->file = (uv_os_fd_t)(ssize_t)-1;
         }
         jl_uv_closeHandle(handle); // synchronous (ok since the callback is known to not interact with any global state)
+        JL_UV_UNLOCK();
         return;
     }
 
-    if (handle->type == UV_NAMED_PIPE || handle->type == UV_TCP) {
-#ifdef _OS_WINDOWS_
-        if (((uv_stream_t*)handle)->stream.conn.shutdown_req) {
-#else
-        if (((uv_stream_t*)handle)->shutdown_req) {
-#endif
-            // don't close the stream while attempting a graceful shutdown
-            return;
-        }
-        if (uv_is_writable((uv_stream_t*)handle)) {
-            // attempt graceful shutdown of writable streams to give them a chance to flush first
-            uv_shutdown_t *req = (uv_shutdown_t*)malloc(sizeof(uv_shutdown_t));
-            req->data = 0;
-            /*
-             * We are explicitly ignoring the error here for the following reason:
-             * There is only two scenarios in which this returns an error:
-             * a) In case the stream is already shut down, in which case we're likely
-             *    in the process of closing this stream (since there's no other call to
-             *    uv_shutdown).
-             * b) In case the stream is already closed, in which case uv_close would
-             *    cause an assertion failure.
-             */
-            uv_shutdown(req, (uv_stream_t*)handle, &jl_uv_shutdownCallback);
-            return;
-        }
+    if (handle->type == UV_NAMED_PIPE || handle->type == UV_TCP || handle->type == UV_TTY) {
+        uv_write_t *req = (uv_write_t*)malloc(sizeof(uv_write_t));
+        req->handle = (uv_stream_t*)handle;
+        jl_uv_flush_close_callback(req, 0);
+        JL_UV_UNLOCK();
+        return;
     }
 
-    if (!uv_is_closing((uv_handle_t*)handle)) {
-        // avoid double-closing the stream
-        if (handle->type == UV_TTY)
-            uv_tty_set_mode((uv_tty_t*)handle, UV_TTY_MODE_NORMAL);
+    // avoid double-closing the stream
+    if (!uv_is_closing(handle)) {
         uv_close(handle, &jl_uv_closeHandle);
     }
+    JL_UV_UNLOCK();
 }
 
 JL_DLLEXPORT void jl_forceclose_uv(uv_handle_t *handle)
 {
-    uv_close(handle, &jl_uv_closeHandle);
+    // avoid double-closing the stream
+    if (!uv_is_closing(handle)) {
+        JL_UV_LOCK();
+        uv_close(handle, &jl_uv_closeHandle);
+        JL_UV_UNLOCK();
+    }
 }
 
 JL_DLLEXPORT void jl_uv_associate_julia_struct(uv_handle_t *handle,
@@ -247,36 +284,40 @@ JL_DLLEXPORT void jl_uv_disassociate_julia_struct(uv_handle_t *handle)
     handle->data = NULL;
 }
 
-JL_DLLEXPORT int jl_spawn(char *name, char **argv, uv_loop_t *loop,
-                          uv_process_t *proc, jl_value_t *julia_struct,
-                          uv_handle_type stdin_type, uv_pipe_t *stdin_pipe,
-                          uv_handle_type stdout_type, uv_pipe_t *stdout_pipe,
-                          uv_handle_type stderr_type, uv_pipe_t *stderr_pipe,
-                          int flags, char **env, char *cwd, uv_exit_cb cb)
+#define UV_CLOSED 0x02 // UV_HANDLE_CLOSED on Windows (same value)
+
+JL_DLLEXPORT int jl_spawn(char *name, char **argv,
+                          uv_loop_t *loop, uv_process_t *proc,
+                          uv_stdio_container_t *stdio, int nstdio,
+                          uint32_t flags, char **env, char *cwd, uv_exit_cb cb)
 {
-    uv_process_options_t opts;
-    uv_stdio_container_t stdio[3];
-    int error;
+    uv_process_options_t opts = {0};
+    opts.stdio = stdio;
     opts.file = name;
     opts.env = env;
-#ifdef _OS_WINDOWS_
     opts.flags = flags;
-#else
-    opts.flags = flags | UV_PROCESS_RESET_SIGPIPE;
-#endif
+    // unused fields:
+    //opts.uid = 0;
+    //opts.gid = 0;
+    //opts.cpumask = NULL;
+    //opts.cpumask_size = 0;
     opts.cwd = cwd;
     opts.args = argv;
-    opts.stdio = stdio;
-    opts.stdio_count = 3;
-    stdio[0].type = stdin_type;
-    stdio[0].data.stream = (uv_stream_t*)(stdin_pipe);
-    stdio[1].type = stdout_type;
-    stdio[1].data.stream = (uv_stream_t*)(stdout_pipe);
-    stdio[2].type = stderr_type;
-    stdio[2].data.stream = (uv_stream_t*)(stderr_pipe);
+    opts.stdio_count = nstdio;
+    while (nstdio--) {
+        int flags = opts.stdio[nstdio].flags;
+        if (!(flags == UV_INHERIT_FD || flags == UV_INHERIT_STREAM || flags == UV_IGNORE)) {
+            proc->type = UV_PROCESS;
+            proc->loop = loop;
+            proc->flags = UV_CLOSED;
+            return UV_EINVAL;
+        }
+    }
     opts.exit_cb = cb;
-    error = uv_spawn(loop,proc,&opts);
-    return error;
+    JL_UV_LOCK();
+    int r = uv_spawn(loop, proc, &opts);
+    JL_UV_UNLOCK();
+    return r;
 }
 
 #ifdef _OS_WINDOWS_
@@ -299,7 +340,7 @@ JL_DLLEXPORT int jl_fs_unlink(char *path)
 {
     uv_fs_t req;
     JL_SIGATOMIC_BEGIN();
-    int ret = uv_fs_unlink(jl_io_loop, &req, path, NULL);
+    int ret = uv_fs_unlink(unused_uv_loop_arg, &req, path, NULL);
     uv_fs_req_cleanup(&req);
     JL_SIGATOMIC_END();
     return ret;
@@ -309,18 +350,18 @@ JL_DLLEXPORT int jl_fs_rename(const char *src_path, const char *dst_path)
 {
     uv_fs_t req;
     JL_SIGATOMIC_BEGIN();
-    int ret = uv_fs_rename(jl_io_loop, &req, src_path, dst_path, NULL);
+    int ret = uv_fs_rename(unused_uv_loop_arg, &req, src_path, dst_path, NULL);
     uv_fs_req_cleanup(&req);
     JL_SIGATOMIC_END();
     return ret;
 }
 
-JL_DLLEXPORT int jl_fs_sendfile(int src_fd, int dst_fd,
+JL_DLLEXPORT int jl_fs_sendfile(uv_os_fd_t src_fd, uv_os_fd_t dst_fd,
                                 int64_t in_offset, size_t len)
 {
     uv_fs_t req;
     JL_SIGATOMIC_BEGIN();
-    int ret = uv_fs_sendfile(jl_io_loop, &req, dst_fd, src_fd,
+    int ret = uv_fs_sendfile(unused_uv_loop_arg, &req, dst_fd, src_fd,
                              in_offset, len, NULL);
     uv_fs_req_cleanup(&req);
     JL_SIGATOMIC_END();
@@ -330,7 +371,7 @@ JL_DLLEXPORT int jl_fs_sendfile(int src_fd, int dst_fd,
 JL_DLLEXPORT int jl_fs_symlink(char *path, char *new_path, int flags)
 {
     uv_fs_t req;
-    int ret = uv_fs_symlink(jl_io_loop, &req, path, new_path, flags, NULL);
+    int ret = uv_fs_symlink(unused_uv_loop_arg, &req, path, new_path, flags, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
@@ -338,7 +379,7 @@ JL_DLLEXPORT int jl_fs_symlink(char *path, char *new_path, int flags)
 JL_DLLEXPORT int jl_fs_chmod(char *path, int mode)
 {
     uv_fs_t req;
-    int ret = uv_fs_chmod(jl_io_loop, &req, path, mode, NULL);
+    int ret = uv_fs_chmod(unused_uv_loop_arg, &req, path, mode, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
@@ -346,45 +387,52 @@ JL_DLLEXPORT int jl_fs_chmod(char *path, int mode)
 JL_DLLEXPORT int jl_fs_chown(char *path, int uid, int gid)
 {
     uv_fs_t req;
-    int ret = uv_fs_chown(jl_io_loop, &req, path, uid, gid, NULL);
+    int ret = uv_fs_chown(unused_uv_loop_arg, &req, path, uid, gid, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
 
-JL_DLLEXPORT int jl_fs_write(int handle, const char *data, size_t len,
+JL_DLLEXPORT int jl_fs_write(uv_os_fd_t handle, const char *data, size_t len,
                              int64_t offset)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
+    // TODO: fix this cheating
     if (ptls->safe_restore || ptls->tid != 0)
+#ifdef _OS_WINDOWS_
+        return WriteFile(handle, data, len, NULL, NULL);
+#else
         return write(handle, data, len);
+#endif
     uv_fs_t req;
     uv_buf_t buf[1];
     buf[0].base = (char*)data;
     buf[0].len = len;
-    int ret = uv_fs_write(jl_io_loop, &req, handle, buf, 1, offset, NULL);
+    if (!jl_io_loop)
+        jl_io_loop = uv_default_loop();
+    int ret = uv_fs_write(unused_uv_loop_arg, &req, handle, buf, 1, offset, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
 
-JL_DLLEXPORT int jl_fs_read(int handle, char *data, size_t len)
+JL_DLLEXPORT int jl_fs_read(uv_os_fd_t handle, char *data, size_t len)
 {
     uv_fs_t req;
     uv_buf_t buf[1];
     buf[0].base = data;
     buf[0].len = len;
-    int ret = uv_fs_read(jl_io_loop, &req, handle, buf, 1, -1, NULL);
+    int ret = uv_fs_read(unused_uv_loop_arg, &req, handle, buf, 1, -1, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
 
-JL_DLLEXPORT int jl_fs_read_byte(int handle)
+JL_DLLEXPORT int jl_fs_read_byte(uv_os_fd_t handle)
 {
     uv_fs_t req;
-    char c;
+    unsigned char c;
     uv_buf_t buf[1];
-    buf[0].base = &c;
+    buf[0].base = (char*)&c;
     buf[0].len = 1;
-    int ret = uv_fs_read(jl_io_loop, &req, handle, buf, 1, -1, NULL);
+    int ret = uv_fs_read(unused_uv_loop_arg, &req, handle, buf, 1, -1, NULL);
     uv_fs_req_cleanup(&req);
     switch (ret) {
     case -1: return ret;
@@ -396,22 +444,24 @@ JL_DLLEXPORT int jl_fs_read_byte(int handle)
     }
 }
 
-JL_DLLEXPORT int jl_fs_close(int handle)
+JL_DLLEXPORT int jl_fs_close(uv_os_fd_t handle)
 {
     uv_fs_t req;
-    int ret = uv_fs_close(jl_io_loop, &req, handle, NULL);
+    int ret = uv_fs_close(unused_uv_loop_arg, &req, handle, NULL);
     uv_fs_req_cleanup(&req);
     return ret;
 }
 
 JL_DLLEXPORT int jl_uv_write(uv_stream_t *stream, const char *data, size_t n,
-                             uv_write_t *uvw, void *writecb)
+                             uv_write_t *uvw, uv_write_cb writecb)
 {
     uv_buf_t buf[1];
     buf[0].base = (char*)data;
     buf[0].len = n;
+    JL_UV_LOCK();
     JL_SIGATOMIC_BEGIN();
-    int err = uv_write(uvw,stream,buf,1,(uv_write_cb)writecb);
+    int err = uv_write(uvw, stream, buf, 1, writecb);
+    JL_UV_UNLOCK();
     JL_SIGATOMIC_END();
     return err;
 }
@@ -432,29 +482,31 @@ JL_DLLEXPORT void jl_uv_puts(uv_stream_t *stream, const char *str, size_t n)
         sizeof(((uv_stream_t*)0)->type) == sizeof(((ios_t*)0)->bm),
             "UV and ios layout mismatch");
 
-    uv_file fd = -1;
+    uv_os_fd_t fd = (uv_os_fd_t)(ssize_t)-1;
 
     // Fallback for output during early initialisation...
-    if (stream == (void*)STDOUT_FILENO || stream == (void*)STDERR_FILENO) {
-        if (!jl_io_loop) jl_io_loop = uv_default_loop();
-        fd = (uv_file)(size_t)stream;
+    if (stream == (void*)STDOUT_FILENO) {
+        fd = UV_STDOUT_FD;
+    }
+    else if (stream == (void*)STDERR_FILENO) {
+        fd = UV_STDERR_FD;
     }
     else if (stream->type == UV_FILE) {
         fd = ((jl_uv_file_t*)stream)->file;
     }
 
-    // Hack to make CoreIO thread-safer
+    // TODO: Hack to make CoreIO thread-safer
     jl_ptls_t ptls = jl_get_ptls_states();
     if (ptls->tid != 0) {
         if (stream == JL_STDOUT) {
-            fd = STDOUT_FILENO;
+            fd = UV_STDOUT_FD;
         }
         else if (stream == JL_STDERR) {
-            fd = STDERR_FILENO;
+            fd = UV_STDERR_FD;
         }
     }
 
-    if (fd != -1) {
+    if ((ssize_t)fd != -1) {
         // Write to file descriptor...
         jl_fs_write(fd, str, n, -1);
     }
@@ -474,8 +526,10 @@ JL_DLLEXPORT void jl_uv_puts(uv_stream_t *stream, const char *str, size_t n)
         buf[0].base = data;
         buf[0].len = n;
         req->data = NULL;
+        JL_UV_LOCK();
         JL_SIGATOMIC_BEGIN();
         int status = uv_write(req, stream, buf, 1, (uv_write_cb)jl_uv_writecb);
+        JL_UV_UNLOCK();
         JL_SIGATOMIC_END();
         if (status < 0) {
             jl_uv_writecb(req, status);
@@ -488,10 +542,21 @@ JL_DLLEXPORT void jl_uv_putb(uv_stream_t *stream, uint8_t b)
     jl_uv_puts(stream, (char*)&b, 1);
 }
 
-JL_DLLEXPORT void jl_uv_putc(uv_stream_t *stream, uint32_t wchar)
+JL_DLLEXPORT void jl_uv_putc(uv_stream_t *stream, uint32_t c)
 {
     char s[4];
-    jl_uv_puts(stream, s, u8_wc_toutf8(s, wchar));
+    int n = 1;
+    s[0] = c >> 24;
+    if ((s[1] = c >> 16)) {
+        n++;
+        if ((s[2] = c >> 8)) {
+            n++;
+            if ((s[3] = c)) {
+                n++;
+            }
+        }
+    }
+    jl_uv_puts(stream, s, n);
 }
 
 extern int vasprintf(char **str, const char *fmt, va_list ap);
@@ -570,6 +635,7 @@ JL_DLLEXPORT int jl_tcp_bind(uv_tcp_t *handle, uint16_t port, uint32_t host,
     addr.sin_port = port;
     addr.sin_addr.s_addr = host;
     addr.sin_family = AF_INET;
+    // TODO: do we need a lock here?
     return uv_tcp_bind(handle, (struct sockaddr*)&addr, flags);
 }
 
@@ -581,6 +647,7 @@ JL_DLLEXPORT int jl_tcp_bind6(uv_tcp_t *handle, uint16_t port, void *host,
     addr.sin6_port = port;
     memcpy(&addr.sin6_addr, host, 16);
     addr.sin6_family = AF_INET6;
+    // TODO: do we need a lock here
     return uv_tcp_bind(handle, (struct sockaddr*)&addr, flags);
 }
 
@@ -669,7 +736,10 @@ JL_DLLEXPORT int jl_udp_send(uv_udp_t *handle, uint16_t port, uint32_t host,
     buf[0].len = size;
     uv_udp_send_t *req = (uv_udp_send_t*)malloc(sizeof(uv_udp_send_t));
     req->data = handle->data;
-    return uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, cb);
+    JL_UV_LOCK();
+    int r = uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, cb);
+    JL_UV_UNLOCK();
+    return r;
 }
 
 JL_DLLEXPORT int jl_udp_send6(uv_udp_t *handle, uint16_t port, void *host,
@@ -685,7 +755,10 @@ JL_DLLEXPORT int jl_udp_send6(uv_udp_t *handle, uint16_t port, void *host,
     buf[0].len = size;
     uv_udp_send_t *req = (uv_udp_send_t *) malloc(sizeof(uv_udp_send_t));
     req->data = handle->data;
-    return uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, cb);
+    JL_UV_LOCK();
+    int r = uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, cb);
+    JL_UV_UNLOCK();
+    return r;
 }
 
 JL_DLLEXPORT int jl_uv_sizeof_interface_address(void)
@@ -709,23 +782,49 @@ JL_DLLEXPORT struct sockaddr_in *jl_uv_interface_address_sockaddr(uv_interface_a
     return &ifa->address.address4;
 }
 
-JL_DLLEXPORT int jl_getaddrinfo(uv_loop_t *loop, const char *host,
-                                const char *service, jl_function_t *cb,
-                                uv_getaddrinfo_cb uvcb)
+JL_DLLEXPORT int jl_getaddrinfo(uv_loop_t *loop, uv_getaddrinfo_t *req,
+        const char *host, const char *service, uv_getaddrinfo_cb uvcb)
 {
-    uv_getaddrinfo_t *req = (uv_getaddrinfo_t*)malloc(sizeof(uv_getaddrinfo_t));
     struct addrinfo hints;
-
-    memset (&hints, 0, sizeof (hints));
+    memset(&hints, 0, sizeof(hints));
     hints.ai_family = PF_UNSPEC;
     hints.ai_socktype = SOCK_STREAM;
     hints.ai_flags |= AI_CANONNAME;
 
-    req->data = cb;
-
-    return uv_getaddrinfo(loop,req,uvcb,host,service,&hints);
+    req->data = NULL;
+    return uv_getaddrinfo(loop, req, uvcb, host, service, &hints);
 }
 
+JL_DLLEXPORT int jl_getnameinfo(uv_loop_t *loop, uv_getnameinfo_t *req,
+        uint32_t host, uint16_t port, int flags, uv_getnameinfo_cb uvcb)
+{
+    struct sockaddr_in addr;
+    memset(&addr, 0, sizeof(addr));
+    addr.sin_family = AF_INET;
+    addr.sin_addr.s_addr = host;
+    addr.sin_port = port;
+
+    req->data = NULL;
+    return uv_getnameinfo(loop, req, uvcb, (struct sockaddr*)&addr, flags);
+}
+
+JL_DLLEXPORT int jl_getnameinfo6(uv_loop_t *loop, uv_getnameinfo_t *req,
+        void *host, uint16_t port, int flags, uv_getnameinfo_cb uvcb)
+{
+    struct sockaddr_in6 addr;
+    memset(&addr, 0, sizeof(addr));
+    addr.sin6_family = AF_INET6;
+    memcpy(&addr.sin6_addr, host, 16);
+    addr.sin6_port = port;
+
+    req->data = NULL;
+    JL_UV_LOCK();
+    int r = uv_getnameinfo(loop, req, uvcb, (struct sockaddr*)&addr, flags);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+
 JL_DLLEXPORT struct sockaddr *jl_sockaddr_from_addrinfo(struct addrinfo *addrinfo)
 {
     return addrinfo->ai_addr;
@@ -785,7 +884,10 @@ JL_DLLEXPORT int jl_tcp4_connect(uv_tcp_t *handle,uint32_t host, uint16_t port,
     addr.sin_family = AF_INET;
     addr.sin_addr.s_addr = host;
     addr.sin_port = port;
-    return uv_tcp_connect(req,handle,(struct sockaddr*)&addr,cb);
+    JL_UV_LOCK();
+    int r = uv_tcp_connect(req,handle,(struct sockaddr*)&addr,cb);
+    JL_UV_UNLOCK();
+    return r;
 }
 
 JL_DLLEXPORT int jl_tcp6_connect(uv_tcp_t *handle, void *host, uint16_t port,
@@ -798,7 +900,10 @@ JL_DLLEXPORT int jl_tcp6_connect(uv_tcp_t *handle, void *host, uint16_t port,
     addr.sin6_family = AF_INET6;
     memcpy(&addr.sin6_addr, host, 16);
     addr.sin6_port = port;
-    return uv_tcp_connect(req,handle,(struct sockaddr*)&addr,cb);
+    JL_UV_LOCK();
+    int r = uv_tcp_connect(req,handle,(struct sockaddr*)&addr,cb);
+    JL_UV_UNLOCK();
+    return r;
 }
 
 JL_DLLEXPORT int jl_connect_raw(uv_tcp_t *handle,struct sockaddr_storage *addr,
@@ -806,7 +911,10 @@ JL_DLLEXPORT int jl_connect_raw(uv_tcp_t *handle,struct sockaddr_storage *addr,
 {
     uv_connect_t *req = (uv_connect_t*)malloc(sizeof(uv_connect_t));
     req->data = 0;
-    return uv_tcp_connect(req,handle,(struct sockaddr*)addr,cb);
+    JL_UV_LOCK();
+    int r = uv_tcp_connect(req,handle,(struct sockaddr*)addr,cb);
+    JL_UV_UNLOCK();
+    return r;
 }
 
 #ifdef _OS_LINUX_
@@ -823,6 +931,15 @@ JL_DLLEXPORT int jl_tcp_quickack(uv_tcp_t *handle, int on)
 
 #endif
 
+JL_DLLEXPORT int jl_has_so_reuseport(void)
+{
+#if defined(SO_REUSEPORT)
+    return 1;
+#else
+    return 0;
+#endif
+}
+
 JL_DLLEXPORT int jl_tcp_reuseport(uv_tcp_t *handle)
 {
 #if defined(SO_REUSEPORT)
@@ -833,7 +950,7 @@ JL_DLLEXPORT int jl_tcp_reuseport(uv_tcp_t *handle)
     }
     return 0;
 #else
-    return 1;
+    return -1;
 #endif
 }
 
@@ -842,12 +959,20 @@ JL_DLLEXPORT int jl_tcp_reuseport(uv_tcp_t *handle)
 JL_DLLEXPORT int jl_uv_unix_fd_is_watched(int fd, uv_poll_t *handle,
                                           uv_loop_t *loop)
 {
-    if (fd >= loop->nwatchers)
+    JL_UV_LOCK();
+    if (fd >= loop->nwatchers) {
+        JL_UV_UNLOCK();
         return 0;
-    if (loop->watchers[fd] == NULL)
+    }
+    if (loop->watchers[fd] == NULL) {
+        JL_UV_UNLOCK();
         return 0;
-    if (handle && loop->watchers[fd] == &handle->io_watcher)
+    }
+    if (handle && loop->watchers[fd] == &handle->io_watcher) {
+        JL_UV_UNLOCK();
         return 0;
+    }
+    JL_UV_UNLOCK();
     return 1;
 }
 
@@ -908,6 +1033,7 @@ JL_DLLEXPORT int jl_tty_set_mode(uv_tty_t *handle, int mode)
     uv_tty_mode_t mode_enum = UV_TTY_MODE_NORMAL;
     if (mode)
         mode_enum = UV_TTY_MODE_RAW;
+    // TODO: do we need lock?
     return uv_tty_set_mode(handle, mode_enum);
 }
 
@@ -941,22 +1067,119 @@ void jl_work_notifier(uv_work_t *req, int status)
     free(baton);
 }
 
-JL_DLLEXPORT int jl_queue_work(void *work_func, void *work_args, void *work_retval,
-                               void *notify_func, int notify_idx)
+JL_DLLEXPORT int jl_queue_work(work_cb_t work_func, void *work_args, void *work_retval,
+                               notify_cb_t notify_func, int notify_idx)
 {
     struct work_baton *baton = (struct work_baton*) malloc(sizeof(struct work_baton));
     baton->req.data = (void*) baton;
-    baton->work_func = (work_cb_t)work_func;
+    baton->work_func = work_func;
     baton->work_args = work_args;
     baton->work_retval = work_retval;
-    baton->notify_func = (notify_cb_t)notify_func;
+    baton->notify_func = notify_func;
     baton->notify_idx = notify_idx;
 
+    JL_UV_LOCK();
     uv_queue_work(jl_io_loop, &baton->req, jl_work_wrapper, jl_work_notifier);
+    JL_UV_UNLOCK();
 
     return 0;
 }
 
+JL_DLLEXPORT void jl_uv_stop(uv_loop_t* loop)
+{
+    JL_UV_LOCK();
+    uv_stop(loop);
+    // TODO: use memory/compiler fence here instead of the lock
+    JL_UV_UNLOCK();
+}
+
+JL_DLLEXPORT void jl_uv_update_time(uv_loop_t* loop)
+{
+    JL_UV_LOCK();
+    uv_update_time(loop);
+    JL_UV_UNLOCK();
+}
+
+JL_DLLEXPORT int jl_uv_timer_start(uv_timer_t* handle, uv_timer_cb cb,
+                                   uint64_t timeout, uint64_t repeat)
+{
+    JL_UV_LOCK();
+    int r = uv_timer_start(handle, cb, timeout, repeat);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_timer_stop(uv_timer_t* handle)
+{
+    JL_UV_LOCK();
+    int r = uv_timer_stop(handle);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_fs_scandir(uv_loop_t* loop, uv_fs_t* req, const char* path, int flags,
+                                  uv_fs_cb cb)
+{
+    JL_UV_LOCK();
+    int r = uv_fs_scandir(loop, req, path, flags, cb);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_fs_readlink(uv_loop_t* loop, uv_fs_t* req, const char* path,
+                                   uv_fs_cb cb)
+{
+    JL_UV_LOCK();
+    int r = uv_fs_readlink(loop, req, path, cb);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_fs_open(uv_loop_t* loop, uv_fs_t* req, const char* path, int flags,
+                               int mode, uv_fs_cb cb)
+{
+    JL_UV_LOCK();
+    int r = uv_fs_open(loop, req, path, flags, mode, cb);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_fs_ftruncate(uv_loop_t* loop, uv_fs_t* req, uv_os_fd_t handle,
+                                    int64_t offset, uv_fs_cb cb)
+{
+    JL_UV_LOCK();
+    int r = uv_fs_ftruncate(loop, req, handle, offset, cb);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_fs_futime(uv_loop_t* loop, uv_fs_t* req, uv_os_fd_t handle, double atime,
+                                 double mtime, uv_fs_cb cb)
+{
+    JL_UV_LOCK();
+    int r = uv_fs_futime(loop, req, handle, atime, mtime, cb);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_read_start(uv_stream_t* handle, uv_alloc_cb alloc_cb,
+                                  uv_read_cb read_cb)
+{
+    JL_UV_LOCK();
+    int r = uv_read_start(handle, alloc_cb, read_cb);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+JL_DLLEXPORT int jl_uv_read_stop(uv_stream_t* handle)
+{
+    JL_UV_LOCK();
+    int r = uv_read_stop(handle);
+    JL_UV_UNLOCK();
+    return r;
+}
+
+
 #ifndef _OS_WINDOWS_
 #if defined(__APPLE__)
 int uv___stream_fd(uv_stream_t *handle);
