@@ -76,7 +76,7 @@ jl_options_t jl_options = { 0,    // quiet
 static const char usage[] = "julia [switches] -- [programfile] [args...]\n";
 static const char opts[]  =
     " -v, --version             Display version information\n"
-    " -h, --help                Print this message\n\n"
+    " -h, --help                Print this message (--help-hidden for more)\n\n"
 
     // startup options
     " --project[={<dir>|@.}]    Set <dir> as the home project/environment\n"
@@ -111,7 +111,6 @@ static const char opts[]  =
     " --warn-overwrite={yes|no} Enable or disable method overwrite warnings\n\n"
 
     // code generation options
-    //" --compile={yes|no|all|min}Enable or disable JIT compiler, or request exhaustive compilation\n"
     " -C, --cpu-target <target> Limit usage of CPU features up to <target>; set to \"help\" to see the available options\n"
     " -O, --optimize={0,1,2,3}  Set the optimization level (default level is 2 if unspecified or 3 if used without a level)\n"
     " -g, -g <level>            Enable / Set the level of debug info generation"
@@ -134,18 +133,22 @@ static const char opts[]  =
     "                           Append coverage information to the LCOV tracefile (filename supports format tokens).\n"
 // TODO: These TOKENS are defined in `runtime_ccall.cpp`. A more verbose `--help` should include that list here.
     " --track-allocation={none|user|all}, --track-allocation\n"
-    "                           Count bytes allocated by each source line\n\n"
+    "                           Count bytes allocated by each source line (omitting setting is equivalent to \"user\")\n\n"
+;
+
+static const char opts_hidden[]  =
+    // code generation options
+    " --compile={yes|no|all|min}Enable or disable JIT compiler, or request exhaustive compilation\n"
 
     // compiler output options
-    //" --output-o name           Generate an object file (including system image data)\n"
-    //" --output-ji name          Generate a system image data file (.ji)\n"
-// These are for compiler debugging purposes only and should not be otherwise
-// used, so don't show them here. See the devdocs for tips on using these
-// options for debugging the compiler.
-//  " --output-unopt-bc name    Generate unoptimized LLVM bitcode (.bc)\n"
-//  " --output-jit-bc name      Dump all IR generated by the frontend (not including system image)\n"
-    //" --output-bc name          Generate LLVM bitcode (.bc)\n"
-    //" --output-incremental=no   Generate an incremental output file (rather than complete)\n\n"
+    " --output-o name           Generate an object file (including system image data)\n"
+    " --output-ji name          Generate a system image data file (.ji)\n"
+
+    // compiler debugging (see the devdocs for tips on using these options)
+    " --output-unopt-bc name    Generate unoptimized LLVM bitcode (.bc)\n"
+    " --output-jit-bc name      Dump all IR generated by the frontend (not including system image)\n"
+    " --output-bc name          Generate LLVM bitcode (.bc)\n"
+    " --output-incremental=no   Generate an incremental output file (rather than complete)\n\n"
 ;
 
 JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
@@ -175,6 +178,7 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
            opt_use_precompiled,
            opt_use_compilecache,
            opt_incremental,
+           opt_help_hidden,
            opt_banner,
            opt_sysimage_native_code,
            opt_compiled_modules,
@@ -188,6 +192,7 @@ JL_DLLEXPORT void jl_parse_opts(int *argcp, char ***argvp)
         // with the required arguments defined in base/client.jl `process_options()`
         { "version",         no_argument,       0, 'v' },
         { "help",            no_argument,       0, 'h' },
+        { "help-hidden",     no_argument,       0, opt_help_hidden },
         { "quiet",           no_argument,       0, 'q' },
         { "banner",          required_argument, 0, opt_banner },
         { "home",            required_argument, 0, 'H' },
@@ -291,10 +296,13 @@ restart_switch:
             break;
         case 'v': // version
             jl_printf(JL_STDOUT, "julia version %s\n", JULIA_VERSION_STRING);
-            jl_exit(0);
+            exit(0);
         case 'h': // help
             jl_printf(JL_STDOUT, "%s%s", usage, opts);
-            jl_exit(0);
+            exit(0);
+        case opt_help_hidden:
+            jl_printf(JL_STDOUT, "%s%s", usage, opts_hidden);
+            exit(0);
         case 'g': // debug info
             if (optarg != NULL) {
                 if (!strcmp(optarg,"0"))
@@ -612,8 +620,9 @@ restart_switch:
     }
     jl_options.code_coverage = codecov;
     jl_options.malloc_log = malloclog;
-    *argvp += optind;
-    *argcp -= optind;
+    int proc_args = *argcp < optind ? *argcp : optind;
+    *argvp += proc_args;
+    *argcp -= proc_args;
 }
 
 JL_DLLEXPORT void jl_set_ARGS(int argc, char **argv)
