@@ -62,7 +62,7 @@ static const jl_fptr_args_t id_to_fptrs[] = {
     jl_f_typeassert, jl_f__apply, jl_f__apply_pure, jl_f__apply_latest, jl_f_isdefined,
     jl_f_tuple, jl_f_svec, jl_f_intrinsic_call, jl_f_invoke_kwsorter,
     jl_f_getfield, jl_f_setfield, jl_f_fieldtype, jl_f_nfields,
-    jl_f_arrayref, jl_f_arrayset, jl_f_arraysize, jl_f_apply_type,
+    jl_f_arrayref, jl_f_const_arrayref, jl_f_arrayset, jl_f_arraysize, jl_f_apply_type,
     jl_f_applicable, jl_f_invoke, jl_f_sizeof, jl_f__expr, jl_f__typevar,
     jl_f_ifelse,
     NULL };
@@ -109,7 +109,7 @@ enum RefTags {
 // calling conventions for internal entry points.
 // this is used to set the method-instance->invoke field
 typedef enum {
-    JL_API_TRAMPOLINE,
+    JL_API_NULL,
     JL_API_BOXED,
     JL_API_CONST,
     JL_API_WITH_PARAMETERS,
@@ -195,9 +195,9 @@ static void jl_load_sysimg_so(void)
     }
     const char *sysimg_data;
     jl_dlsym(jl_sysimg_handle, "jl_system_image_data", (void **)&sysimg_data, 1);
-    size_t len;
-    jl_dlsym(jl_sysimg_handle, "jl_system_image_size", (void **)&len, 1);
-    jl_restore_system_image_data(sysimg_data, len);
+    size_t *plen;
+    jl_dlsym(jl_sysimg_handle, "jl_system_image_size", (void **)&plen, 1);
+    jl_restore_system_image_data(sysimg_data, *plen);
 }
 
 
@@ -673,21 +673,21 @@ static void jl_write_values(jl_serializer_state *s)
             if (jl_is_method(v)) {
                 write_padding(s->s, sizeof(jl_method_t) - tot);
             }
-            else if (jl_is_method_instance(v)) {
-                jl_method_instance_t *m = (jl_method_instance_t*)v;
-                jl_method_instance_t *newm = (jl_method_instance_t*)&s->s->buf[reloc_offset];
+            else if (jl_is_code_instance(v)) {
+                jl_code_instance_t *m = (jl_code_instance_t*)v;
+                jl_code_instance_t *newm = (jl_code_instance_t*)&s->s->buf[reloc_offset];
 
                 newm->invoke = NULL;
                 newm->specptr.fptr = NULL;
                 newm->functionObjectsDecls.functionObject = NULL;
                 newm->functionObjectsDecls.specFunctionObject = NULL;
-                uintptr_t fptr_id = JL_API_TRAMPOLINE;
+                uintptr_t fptr_id = JL_API_NULL;
                 uintptr_t specfptr_id = 0;
                 if (m->invoke == jl_fptr_const_return) {
                     fptr_id = JL_API_CONST;
                 }
                 else {
-                    if (jl_is_method(m->def.method)) {
+                    if (jl_is_method(m->def->def.method)) {
                         specfptr_id = jl_fptr_id(m->specptr.fptr);
                         const char *fname = m->functionObjectsDecls.functionObject;
                         if (specfptr_id) { // found in the table of builtins
@@ -727,10 +727,12 @@ static void jl_write_values(jl_serializer_state *s)
                     }
                 }
                 newm->invoke = NULL; // relocation offset
-                arraylist_push(&s->relocs_list, (void*)(reloc_offset + offsetof(jl_method_instance_t, invoke))); // relocation location
-                arraylist_push(&s->relocs_list, (void*)(((uintptr_t)FunctionRef << RELOC_TAG_OFFSET) + fptr_id)); // relocation target
+                if (fptr_id != JL_API_NULL) {
+                    arraylist_push(&s->relocs_list, (void*)(reloc_offset + offsetof(jl_code_instance_t, invoke))); // relocation location
+                    arraylist_push(&s->relocs_list, (void*)(((uintptr_t)FunctionRef << RELOC_TAG_OFFSET) + fptr_id)); // relocation target
+                }
                 if (specfptr_id >= 2) {
-                    arraylist_push(&s->relocs_list, (void*)(reloc_offset + offsetof(jl_method_instance_t, specptr.fptr))); // relocation location
+                    arraylist_push(&s->relocs_list, (void*)(reloc_offset + offsetof(jl_code_instance_t, specptr.fptr))); // relocation location
                     arraylist_push(&s->relocs_list, (void*)(((uintptr_t)BuiltinFunctionRef << RELOC_TAG_OFFSET) + specfptr_id - 2)); // relocation target
                 }
             }
@@ -895,15 +897,14 @@ static inline uintptr_t get_item_for_reloc(jl_serializer_state *s, uintptr_t bas
         case JL_API_WITH_PARAMETERS:
             if (sysimg_fptrs.base)
                 return (uintptr_t)jl_fptr_sparam;
-            JL_FALLTHROUGH;
-        case JL_API_TRAMPOLINE:
-            return (uintptr_t)jl_fptr_trampoline;
+            return (uintptr_t)NULL;
         case JL_API_CONST:
             return (uintptr_t)jl_fptr_const_return;
         case JL_API_INTERPRETED:
             return (uintptr_t)jl_fptr_interpret_call;
         case JL_API_BUILTIN:
             return (uintptr_t)jl_fptr_args;
+        case JL_API_NULL:
         case JL_API_MAX:
         //default:
             assert("corrupt relocation item id");
@@ -1024,11 +1025,11 @@ static void jl_update_all_fptrs(jl_serializer_state *s)
                 specfunc = 0;
                 offset = ~offset;
             }
-            jl_method_instance_t *li = (jl_method_instance_t*)(base + offset);
+            jl_code_instance_t *codeinst = (jl_code_instance_t*)(base + offset);
             uintptr_t base = (uintptr_t)fvars.base;
-            assert(jl_is_method(li->def.method) && li->invoke != jl_fptr_const_return);
-            assert(specfunc ? li->invoke != jl_fptr_trampoline : li->invoke == jl_fptr_trampoline);
-            linfos[i] = li;
+            assert(jl_is_method(codeinst->def->def.method) && codeinst->invoke != jl_fptr_const_return);
+            assert(specfunc ? codeinst->invoke != NULL : codeinst->invoke == NULL);
+            linfos[i] = codeinst->def;
             int32_t offset = fvars.offsets[i];
             for (; clone_idx < fvars.nclones; clone_idx++) {
                 uint32_t idx = fvars.clone_idxs[clone_idx] & jl_sysimg_val_mask;
@@ -1040,10 +1041,10 @@ static void jl_update_all_fptrs(jl_serializer_state *s)
             }
             void *fptr = (void*)(base + offset);
             if (specfunc)
-                li->specptr.fptr = fptr;
+                codeinst->specptr.fptr = fptr;
             else
-                li->invoke = (jl_callptr_t)fptr;
-            jl_fptr_to_llvm(fptr, li, specfunc);
+                codeinst->invoke = (jl_callptr_t)fptr;
+            jl_fptr_to_llvm(fptr, codeinst, specfunc);
         }
     }
     jl_register_fptrs(sysimage_base, &fvars, linfos, sysimg_fvars_max);
@@ -1629,10 +1630,10 @@ static void jl_init_serializer2(int for_serialize)
                      jl_symbol_type, jl_ssavalue_type, jl_datatype_type, jl_slotnumber_type,
                      jl_simplevector_type, jl_array_type, jl_typedslot_type,
                      jl_expr_type, jl_globalref_type, jl_string_type,
-                     jl_module_type, jl_tvar_type, jl_method_instance_type, jl_method_type,
+                     jl_module_type, jl_tvar_type, jl_method_instance_type, jl_method_type, jl_code_instance_type,
                      jl_emptysvec, jl_emptytuple, jl_false, jl_true, jl_nothing, jl_any_type,
                      call_sym, invoke_sym, goto_ifnot_sym, return_sym, jl_symbol("tuple"),
-                     unreachable_sym,
+                     unreachable_sym, jl_an_empty_string,
                      jl_linenumbernode_type, jl_lineinfonode_type,
                      jl_gotonode_type, jl_quotenode_type,
                      jl_pinode_type, jl_phinode_type, jl_phicnode_type, jl_upsilonnode_type,
@@ -1654,7 +1655,7 @@ static void jl_init_serializer2(int for_serialize)
                      jl_methtable_type->name, jl_typemap_level_type->name, jl_typemap_entry_type->name, jl_tvar_type->name,
                      ((jl_datatype_t*)jl_unwrap_unionall((jl_value_t*)jl_abstractarray_type))->name,
                      ((jl_datatype_t*)jl_unwrap_unionall((jl_value_t*)jl_densearray_type))->name,
-                     jl_vararg_typename, jl_void_type->name, jl_method_instance_type->name, jl_method_type->name,
+                     jl_vararg_typename, jl_void_type->name, jl_method_instance_type->name, jl_method_type->name, jl_code_instance_type->name,
                      jl_module_type->name, jl_function_type->name, jl_typedslot_type->name,
                      jl_abstractslot_type->name, jl_slotnumber_type->name,
                      jl_unionall_type->name, jl_intrinsic_type->name, jl_task_type->name,
