@@ -27,6 +27,7 @@
 
 #include "utils.h"
 #include "utf8.h"
+#include "utf8proc.h"
 #include "ios.h"
 #include "timefuncs.h"
 
@@ -566,7 +567,7 @@ int ios_trunc(ios_t *s, size_t size)
 #if !defined(_OS_WINDOWS_)
         if (ftruncate(s->fd, size) == 0)
 #else
-        if (_chsize(s->fd, size) == 0)
+        if (_chsize_s(s->fd, size) == 0)
 #endif
             return 0;
     }
@@ -858,6 +859,7 @@ static void _ios_init(ios_t *s)
     s->ndirty = 0;
     s->fpos = -1;
     s->lineno = 1;
+    s->u_colno = 0;
     s->fd = -1;
     s->ownbuf = 1;
     s->ownfd = 0;
@@ -870,19 +872,30 @@ static void _ios_init(ios_t *s)
 /* stream object initializers. we do no allocation. */
 
 #if !defined(_OS_WINDOWS_)
+/*
+ * NOTE: we do not handle system call restart in this function,
+ * please do it manually:
+ *
+ *  do
+ *      open_cloexec(...)
+ *  while (-1 == fd && _enonfatal(errno))
+ */
 static int open_cloexec(const char *path, int flags, mode_t mode)
 {
 #ifdef O_CLOEXEC
-    static int no_cloexec=0;
+    static int no_cloexec = 0;
 
     if (!no_cloexec) {
         set_io_wait_begin(1);
         int fd = open(path, flags | O_CLOEXEC, mode);
         set_io_wait_begin(0);
+
         if (fd != -1)
             return fd;
         if (errno != EINVAL)
             return -1;
+
+        /* O_CLOEXEC not supported. */
         no_cloexec = 1;
     }
 #endif
@@ -914,12 +927,15 @@ ios_t *ios_file(ios_t *s, const char *fname, int rd, int wr, int create, int tru
 #else
     // The mode of the created file is (mode & ~umask), which resolves with
     // default umask to u=rw,g=r,o=r
-    fd = open_cloexec(fname, flags,
-                      S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+    do
+        fd = open_cloexec(fname, flags,
+                          S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+    while (-1 == fd && _enonfatal(errno));
 #endif
-    s = ios_fd(s, fd, 1, 1);
     if (fd == -1)
         goto open_file_err;
+
+    s = ios_fd(s, fd, 1, 1);
     if (!rd)
         s->readable = 0;
     if (!wr)
@@ -1093,6 +1109,10 @@ int ios_getutf8(ios_t *s, uint32_t *pwc)
     c0 = (char)c;
     if ((unsigned char)c0 < 0x80) {
         *pwc = (uint32_t)(unsigned char)c0;
+        if (c == '\n')
+            s->u_colno = 0;
+        else
+            s->u_colno += utf8proc_charwidth(*pwc);
         return 1;
     }
     sz = u8_seqlen(&c0);
@@ -1103,6 +1123,7 @@ int ios_getutf8(ios_t *s, uint32_t *pwc)
         return IOS_EOF;
     size_t i = s->bpos;
     *pwc = u8_nextchar(s->buf, &i);
+    s->u_colno += utf8proc_charwidth(*pwc);
     ios_read(s, buf, sz);
     return 1;
 }
