@@ -49,7 +49,7 @@ volatile int jl_in_stackwalk = 0;
 #endif
 #endif
 
-// empirically, finish_task needs about 64k stack space to infer/run
+// empirically, jl_finish_task needs about 64k stack space to infer/run
 // and additionally, gc-stack reserves 64k for the guard pages
 #if defined(MINSIGSTKSZ) && MINSIGSTKSZ > 131072
 #define MINSTKSZ MINSIGSTKSZ
@@ -59,9 +59,9 @@ volatile int jl_in_stackwalk = 0;
 
 #define ROOT_TASK_STACK_ADJUSTMENT 3000000
 
-static jl_sym_t *done_sym;
-static jl_sym_t *failed_sym;
-static jl_sym_t *runnable_sym;
+jl_sym_t *done_sym;
+jl_sym_t *failed_sym;
+jl_sym_t *runnable_sym;
 
 extern size_t jl_page_size;
 jl_datatype_t *jl_task_type;
@@ -74,6 +74,15 @@ static void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t);
 static JL_THREAD unw_cursor_t jl_basecursor;
 #endif
 
+#ifdef ALWAYS_COPY_STACKS
+# ifndef COPY_STACKS
+# error "ALWAYS_COPY_STACKS requires COPY_STACKS"
+# endif
+static int always_copy_stacks = 1;
+#else
+static int always_copy_stacks = 0;
+#endif
+
 #ifdef COPY_STACKS
 
 static void memcpy_a16(uint64_t *to, uint64_t *from, size_t nb)
@@ -101,6 +110,7 @@ static void NOINLINE save_stack(jl_ptls_t ptls, jl_task_t *lastt, jl_task_t **pt
     }
     *pt = lastt; // clear the gc-root for the target task before copying the stack for saving
     lastt->copy_stack = nb;
+    lastt->sticky = 1;
     memcpy_a16((uint64_t*)buf, (uint64_t*)frame_addr, nb);
     // this task's stack could have been modified after
     // it was marked by an incremental collection
@@ -139,7 +149,7 @@ static void restore_stack2(jl_ptls_t ptls, jl_task_t *lastt)
 
 static jl_function_t *task_done_hook_func = NULL;
 
-static void JL_NORETURN finish_task(jl_task_t *t, jl_value_t *resultval JL_MAYBE_UNROOTED)
+void JL_NORETURN jl_finish_task(jl_task_t *t, jl_value_t *resultval JL_MAYBE_UNROOTED)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     JL_SIGATOMIC_BEGIN();
@@ -155,15 +165,7 @@ static void JL_NORETURN finish_task(jl_task_t *t, jl_value_t *resultval JL_MAYBE
     ptls->in_finalizer = 0;
     ptls->in_pure_callback = 0;
     jl_get_ptls_states()->world_age = jl_world_counter;
-    if (ptls->tid != 0) {
-        // For now, only thread 0 runs the task scheduler.
-        // The others return to the thread loop
-        ptls->root_task->result = jl_nothing;
-        jl_task_t *task = ptls->root_task;
-        jl_switchto(&task);
-        gc_debug_critical_error();
-        abort();
-    }
+    // let the runtime know this task is dead and find a new task to run
     if (task_done_hook_func == NULL) {
         task_done_hook_func = (jl_function_t*)jl_get_global(jl_base_module,
                                                             jl_symbol("task_done_hook"));
@@ -254,6 +256,7 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
 #ifdef COPY_STACKS
                 // fall back to stack copying if mmap fails
                 t->copy_stack = 1;
+                t->sticky = 1;
                 t->bufsz = 0;
                 memcpy(&t->ctx, &ptls->base_ctx, sizeof(t->ctx));
 #else
@@ -276,6 +279,7 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
         if (lastt->copy_stack) { // save the old copy-stack
             save_stack(ptls, lastt, pt); // allocates (gc-safepoint, and can also fail)
             if (jl_setjmp(lastt->ctx.uc_mcontext, 0)) {
+                // TODO: mutex unlock the thread we just switched from
 #ifdef ENABLE_TIMINGS
                 assert(blk == ptls->current_task->timing_stack);
                 if (blk)
@@ -296,6 +300,10 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
     ptls->world_age = t->world_age;
     t->gcstack = NULL;
     ptls->current_task = t;
+    if (!lastt->sticky)
+        // release lastt to run on any tid
+        lastt->tid = -1;
+    t->tid = ptls->tid;
 
     jl_ucontext_t *lastt_ctx = (killed ? NULL : &lastt->ctx);
 #ifdef COPY_STACKS
@@ -324,8 +332,12 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
             jl_swap_fiber(lastt_ctx, &t->ctx);
     }
     else {
-        jl_start_fiber(lastt_ctx, &t->ctx);
+        if (always_copy_stacks)
+            jl_longjmp(ptls->base_ctx.uc_mcontext, 1);
+        else
+            jl_start_fiber(lastt_ctx, &t->ctx);
     }
+    // TODO: mutex unlock the thread we just switched from
 #ifdef ENABLE_TIMINGS
     assert(blk == ptls->current_task->timing_stack);
     if (blk)
@@ -452,19 +464,20 @@ JL_DLLEXPORT void jl_rethrow_other(jl_value_t *e JL_MAYBE_UNROOTED)
     throw_internal(NULL);
 }
 
-JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, size_t ssize)
+JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion_future, size_t ssize)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     jl_task_t *t = (jl_task_t*)jl_gc_alloc(ptls, sizeof(jl_task_t), jl_task_type);
     t->copy_stack = 0;
     if (ssize == 0) {
         // stack size unspecified; use default
-#if defined(COPY_STACKS) && defined(ALWAYS_COPY_STACKS)
-        t->copy_stack = 1;
-        t->bufsz = 0;
-#else
-        t->bufsz = JL_STACK_SIZE;
-#endif
+        if (always_copy_stacks) {
+            t->copy_stack = 1;
+            t->bufsz = 0;
+        }
+        else {
+            t->bufsz = JL_STACK_SIZE;
+        }
     }
     else {
         // user requested dedicated stack of a certain size
@@ -475,23 +488,26 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, size_t ssize)
         if (t->stkbuf == NULL)
             jl_throw(jl_memory_exception);
     }
+    t->next = jl_nothing;
+    t->queue = jl_nothing;
     t->tls = jl_nothing;
     t->state = runnable_sym;
     t->start = start;
     t->result = jl_nothing;
-    t->donenotify = jl_nothing;
+    t->donenotify = completion_future;
     t->exception = jl_nothing;
     t->backtrace = jl_nothing;
     // Inherit logger state from parent task
     t->logstate = ptls->current_task->logstate;
     // there is no active exception handler available on this stack yet
     t->eh = NULL;
-    t->tid = 0;
+    t->sticky = 1;
     t->gcstack = NULL;
     t->excstack = NULL;
     t->stkbuf = NULL;
-    t->tid = 0;
     t->started = 0;
+    t->prio = -1;
+    t->tid = 0;
 #ifdef ENABLE_TIMINGS
     t->timing_stack = NULL;
 #endif
@@ -524,7 +540,9 @@ void jl_init_tasks(void) JL_GC_DISABLED
                         NULL,
                         jl_any_type,
                         jl_emptysvec,
-                        jl_perm_symsvec(8,
+                        jl_perm_symsvec(11,
+                                        "next",
+                                        "queue",
                                         "storage",
                                         "state",
                                         "donenotify",
@@ -532,8 +550,11 @@ void jl_init_tasks(void) JL_GC_DISABLED
                                         "exception",
                                         "backtrace",
                                         "logstate",
-                                        "code"),
-                        jl_svec(8,
+                                        "code",
+                                        "sticky"),
+                        jl_svec(11,
+                                jl_any_type,
+                                jl_any_type,
                                 jl_any_type,
                                 jl_sym_type,
                                 jl_any_type,
@@ -541,8 +562,11 @@ void jl_init_tasks(void) JL_GC_DISABLED
                                 jl_any_type,
                                 jl_any_type,
                                 jl_any_type,
-                                jl_any_type),
-                        0, 1, 7);
+                                jl_any_type,
+                                jl_bool_type),
+                        0, 1, 9);
+    jl_value_t *listt = jl_new_struct(jl_uniontype_type, jl_task_type, jl_void_type);
+    jl_svecset(jl_task_type->types, 0, listt);
     done_sym = jl_symbol("done");
     failed_sym = jl_symbol("failed");
     runnable_sym = jl_symbol("runnable");
@@ -579,7 +603,7 @@ static void NOINLINE JL_NORETURN start_task(void)
         }
 skip_pop_exception:;
     }
-    finish_task(t, res);
+    jl_finish_task(t, res);
     gc_debug_critical_error();
     abort();
 }
@@ -767,6 +791,29 @@ static void jl_start_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t)
                     // because all our addresses are word-aligned.
         " udf #0" // abort
         : : "r" (stk), "r"(fn) : "memory" );
+#elif defined(_CPU_PPC64_)
+    // N.B.: There is two iterations of the PPC64 ABI.
+    // v2 is current and used here. Make sure you have the
+    // correct version of the ABI reference when working on this code.
+    asm volatile(
+        // Move stack (-0x30 for initial stack frame) to stack pointer
+        " addi 1, %0, -0x30;\n"
+        // Build stack frame
+        // Skip local variable save area
+        " std 2, 0x28(1);\n" // Save TOC
+        // Clear link editor/compiler words
+        " std 0, 0x20(1);\n"
+        " std 0, 0x18(1);\n"
+        // Clear LR/CR save area
+        " std 0, 0x10(1);\n"
+        " std 0, 0x8(1);\n"
+        " std 0, 0x0(1); \n" // Clear back link to terminate unwinder
+        " mtlr 0; \n"        // Clear link register
+        " mr 12, %1; \n"     // Set up target global entry point
+        " mtctr 12; \n"      // Move jump target to counter register
+        " bctr; \n"          // branch to counter (lr update disabled)
+        " trap; \n"
+        : : "r"(stk), "r"(fn) : "memory");
 #else
 #error JL_HAVE_ASM defined but not implemented for this CPU type
 #endif
@@ -888,7 +935,6 @@ void jl_init_root_task(void *stack_lo, void *stack_hi)
     jl_ptls_t ptls = jl_get_ptls_states();
     ptls->current_task = (jl_task_t*)jl_gc_alloc(ptls, sizeof(jl_task_t),
                                                  jl_task_type);
-    ptls->current_task->copy_stack = 0;
     void *stack = stack_lo;
     size_t ssize = (char*)stack_hi - (char*)stack_lo;
 #ifndef _OS_WINDOWS_
@@ -897,9 +943,19 @@ void jl_init_root_task(void *stack_lo, void *stack_hi)
         ssize += ROOT_TASK_STACK_ADJUSTMENT; // sizeof stack is known exactly, but not where we are in that stack
     }
 #endif
-    ptls->current_task->stkbuf = stack;
-    ptls->current_task->bufsz = ssize;
+    if (always_copy_stacks) {
+        ptls->current_task->copy_stack = 1;
+        ptls->current_task->stkbuf = NULL;
+        ptls->current_task->bufsz = 0;
+    }
+    else {
+        ptls->current_task->copy_stack = 0;
+        ptls->current_task->stkbuf = stack;
+        ptls->current_task->bufsz = ssize;
+    }
     ptls->current_task->started = 1;
+    ptls->current_task->next = jl_nothing;
+    ptls->current_task->queue = jl_nothing;
     ptls->current_task->tls = jl_nothing;
     ptls->current_task->state = runnable_sym;
     ptls->current_task->start = NULL;
@@ -912,13 +968,22 @@ void jl_init_root_task(void *stack_lo, void *stack_hi)
     ptls->current_task->gcstack = NULL;
     ptls->current_task->excstack = NULL;
     ptls->current_task->tid = ptls->tid;
+    ptls->current_task->sticky = 1;
 #ifdef JULIA_ENABLE_THREADING
     arraylist_new(&ptls->current_task->locks, 0);
 #endif
 
     ptls->root_task = ptls->current_task;
 
-    jl_init_basefiber(JL_STACK_SIZE);
+    if (always_copy_stacks) {
+        ptls->stackbase = stack_hi;
+        ptls->stacksize = ssize;
+        if (jl_setjmp(ptls->base_ctx.uc_mcontext, 0))
+            start_task();
+    }
+    else {
+        jl_init_basefiber(JL_STACK_SIZE);
+    }
 }
 
 JL_DLLEXPORT int jl_is_task_started(jl_task_t *t)
@@ -926,6 +991,12 @@ JL_DLLEXPORT int jl_is_task_started(jl_task_t *t)
     return t->started;
 }
 
+JL_DLLEXPORT int16_t jl_get_task_tid(jl_task_t *t)
+{
+    return t->tid;
+}
+
+
 #ifdef _OS_WINDOWS_
 #if defined(_CPU_X86_)
 extern DWORD32 __readgsdword(int);
