@@ -26,7 +26,7 @@ let err = nothing
     end
 end
 
-if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
+if !Sys.iswindows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
     dirlink = joinpath(dir, "dirlink")
     symlink(subdir, dirlink)
     # relative link
@@ -37,7 +37,7 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
     cd(pwd_)
 end
 
-if !is_windows()
+if !Sys.iswindows()
     link = joinpath(dir, "afilelink.txt")
     symlink(file, link)
     # relative link
@@ -52,6 +52,7 @@ end
 #######################################################################
 # This section tests some of the features of the stat-based file info #
 #######################################################################
+@test !isfile(Base.Filesystem.StatStruct())
 @test isdir(dir)
 @test !isfile(dir)
 @test !islink(dir)
@@ -61,13 +62,13 @@ end
 
 @test filemode(file) & 0o444 > 0 # readable
 @test filemode(file) & 0o222 > 0 # writable
-chmod(file, filemode(file) & 0o7555)
+@test chmod(file, filemode(file) & 0o7555) == file
 @test filemode(file) & 0o222 == 0
 chmod(file, filemode(file) | 0o222)
 @test filemode(file) & 0o111 == 0
 @test filesize(file) == 0
 
-if is_windows()
+if Sys.iswindows()
     permissions = 0o444
     @test filemode(dir) & 0o777 != permissions
     @test filemode(subdir) & 0o777 != permissions
@@ -78,10 +79,28 @@ if is_windows()
     @test filemode(file) & 0o777 == permissions
     chmod(dir, 0o666, recursive=true)  # Reset permissions in case someone wants to use these later
 else
+    function get_umask()
+        umask = ccall(:umask, UInt32, (UInt32,), 0)
+        ccall(:umask, UInt32, (UInt32,), umask)
+        return umask
+    end
+
     mktempdir() do tmpdir
+        umask = get_umask()
         tmpfile=joinpath(tmpdir, "tempfile.txt")
+        tmpfile2=joinpath(tmpdir, "tempfile2.txt")
         touch(tmpfile)
+        cp(tmpfile, tmpfile2)
+        @test filemode(tmpfile) & (~umask) == filemode(tmpfile2)
+        rm(tmpfile2)
+        chmod(tmpfile, 0o777)
+        cp(tmpfile, tmpfile2)
+        @test filemode(tmpfile) & (~umask) == filemode(tmpfile2)
+        rm(tmpfile2)
         chmod(tmpfile, 0o707)
+        cp(tmpfile, tmpfile2)
+        @test filemode(tmpfile) & (~umask) == filemode(tmpfile2)
+        rm(tmpfile2)
         linkfile=joinpath(dir, "tempfile.txt")
         symlink(tmpfile, linkfile)
         permissions=0o776
@@ -103,7 +122,7 @@ end
 
 # On windows the filesize of a folder is the accumulation of all the contained
 # files and is thus zero in this case.
-if is_windows()
+if Sys.iswindows()
     @test filesize(dir) == 0
 else
     @test filesize(dir) > 0
@@ -118,15 +137,15 @@ end
 #@test Int(time()) >= Int(mtime(file)) >= Int(mtime(dir)) >= 0 # 1 second accuracy should be sufficient
 
 # test links
-if is_unix()
+if Sys.isunix()
     @test islink(link) == true
     @test readlink(link) == file
 end
 
-if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
+if !Sys.iswindows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
     @test islink(dirlink) == true
     @test isdir(dirlink) == true
-    @test readlink(dirlink) == subdir * (is_windows() ? "\\" : "")
+    @test readlink(dirlink) == subdir * (Sys.iswindows() ? "\\" : "")
 end
 
 # rm recursive TODO add links
@@ -157,12 +176,12 @@ close(f)
 
 rm(c_tmpdir, recursive=true)
 @test !isdir(c_tmpdir)
-@test_throws Base.UVError rm(c_tmpdir)
+@test_throws Base.IOError rm(c_tmpdir)
 @test rm(c_tmpdir, force=true) === nothing
-@test_throws Base.UVError rm(c_tmpdir, recursive=true)
+@test_throws Base.IOError rm(c_tmpdir, recursive=true)
 @test rm(c_tmpdir, force=true, recursive=true) === nothing
 
-if !is_windows()
+if !Sys.iswindows()
     # chown will give an error if the user does not have permissions to change files
     if get(ENV, "USER", "") == "root" || get(ENV, "HOME", "") == "/root"
         chown(file, -2, -1)  # Change the file owner to nobody
@@ -170,106 +189,18 @@ if !is_windows()
         chown(file, 0, -2)  # Change the file group to nogroup (and owner back to root)
         @test stat(file).gid !=0
         @test stat(file).uid ==0
-        chown(file, -1, 0)
+        @test chown(file, -1, 0) == file
         @test stat(file).gid ==0
         @test stat(file).uid ==0
     else
-        @test_throws Base.UVError chown(file, -2, -1)  # Non-root user cannot change ownership to another user
-        @test_throws Base.UVError chown(file, -1, -2)  # Non-root user cannot change group to a group they are not a member of (eg: nogroup)
+        @test_throws Base.IOError chown(file, -2, -1)  # Non-root user cannot change ownership to another user
+        @test_throws Base.IOError chown(file, -1, -2)  # Non-root user cannot change group to a group they are not a member of (eg: nogroup)
     end
 else
     # test that chown doesn't cause any errors for Windows
-    @test chown(file, -2, -2) === nothing
+    @test chown(file, -2, -2) == file
 end
 
-#######################################################################
-# This section tests file watchers.                                   #
-#######################################################################
-function test_file_poll(channel,interval,timeout_s)
-    rc = poll_file(file, interval, timeout_s)
-    put!(channel,rc)
-end
-
-function test_timeout(tval)
-    tic()
-    channel = Channel(1)
-    @async test_file_poll(channel, 10, tval)
-    tr = take!(channel)
-    t_elapsed = toq()
-    @test !ispath(tr[1]) && !ispath(tr[2])
-    @test tval <= t_elapsed
-end
-
-function test_touch(slval)
-    tval = slval*1.1
-    channel = Channel(1)
-    @async test_file_poll(channel, tval/3, tval)
-    sleep(tval/3)  # one poll period
-    f = open(file,"a")
-    write(f,"Hello World\n")
-    close(f)
-    tr = take!(channel)
-    @test ispath(tr[1]) && ispath(tr[2])
-end
-
-function test_watch_file_timeout(tval)
-    watch = @async watch_file(file, tval)
-    @test wait(watch) == Base.Filesystem.FileEvent(false, false, true)
-end
-
-function test_watch_file_change(tval)
-    watch = @async watch_file(file, tval)
-    sleep(tval/3)
-    open(file, "a") do f
-        write(f, "small change\n")
-    end
-    @test wait(watch) == Base.Filesystem.FileEvent(false, true, false)
-end
-
-function test_monitor_wait(tval)
-    fm = FileMonitor(file)
-    @async begin
-        sleep(tval)
-        f = open(file,"a")
-        write(f,"Hello World\n")
-        close(f)
-    end
-    fname, events = wait(fm)
-    close(fm)
-    if is_linux() || is_windows() || is_apple()
-        @test fname == basename(file)
-    else
-        @test fname == ""  # platforms where F_GETPATH is not available
-    end
-    @test events.changed
-end
-
-function test_monitor_wait_poll()
-    pfw = PollingFileWatcher(file, 5.007)
-    @async begin
-        sleep(2.5)
-        f = open(file,"a")
-        write(f,"Hello World\n")
-        close(f)
-    end
-    (old, new) = wait(pfw)
-    close(pfw)
-    @test new.mtime - old.mtime > 2.5 - 1.5 # mtime may only have second-level accuracy (plus add some hysteresis)
-end
-
-test_timeout(0.1)
-test_timeout(1)
-test_touch(6)
-test_monitor_wait(0.1)
-test_monitor_wait(0.1)
-test_monitor_wait_poll()
-test_monitor_wait_poll()
-test_watch_file_timeout(0.1)
-test_watch_file_change(6)
-
-@test_throws Base.UVError watch_file("____nonexistent_file", 10)
-@test_throws Base.UVError poll_file("____nonexistent_file", 2, 10)
-
 ##############
 # mark/reset #
 ##############
@@ -302,19 +233,30 @@ close(s)
 # This section tests temporary file and directory creation.           #
 #######################################################################
 
+@testset "quoting filenames" begin
+    @test try
+        open("this file is not expected to exist")
+        false
+    catch e
+        isa(e, SystemError) || rethrow()
+        @test sprint(showerror, e) == "SystemError: opening file \"this file is not expected to exist\": No such file or directory"
+        true
+    end
+end
+
 my_tempdir = tempdir()
 @test isdir(my_tempdir) == true
 
-path = tempname()
-# Issue #9053.
-@test ispath(path) == is_windows()
-ispath(path) && rm(path)
+let path = tempname()
+    # issue #9053
+    @test !ispath(path)
+end
 
 (p, f) = mktemp()
 print(f, "Here is some text")
 close(f)
 @test isfile(p) == true
-@test readstring(p) == "Here is some text"
+@test read(p, String) == "Here is some text"
 rm(p)
 
 let
@@ -339,7 +281,7 @@ end
 emptyfile = joinpath(dir, "empty")
 touch(emptyfile)
 emptyf = open(emptyfile)
-@test isempty(readlines(emptyf, chomp=false))
+@test isempty(readlines(emptyf, keep=true))
 close(emptyf)
 rm(emptyfile)
 
@@ -382,7 +324,7 @@ function check_cp(orig_path::AbstractString, copied_path::AbstractString, follow
                 # copied_path must also be a file.
                 @test isfile(copied_path)
                 # copied_path must have same content
-                @test readstring(orig_path) == readstring(copied_path)
+                @test read(orig_path, String) == read(copied_path, String)
             end
         end
     elseif isdir(orig_path)
@@ -391,7 +333,7 @@ function check_cp(orig_path::AbstractString, copied_path::AbstractString, follow
         # copied_path must also be a file.
         @test isfile(copied_path)
         # copied_path must have same content
-        @test readstring(orig_path) == readstring(copied_path)
+        @test read(orig_path, String) == read(copied_path, String)
     end
 end
 
@@ -421,7 +363,7 @@ cp(afile, bfile)
 
 cfile = joinpath(dir, "c.txt")
 write(cfile, "This is longer than the contents of afile")
-cp(afile, cfile; remove_destination=true)
+cp(afile, cfile; force=true)
 
 a_stat = stat(afile)
 b_stat = stat(bfile)
@@ -476,7 +418,7 @@ end
 
 # issue #10506 #10434
 ## Tests for directories and links to directories
-if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
+if !Sys.iswindows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
     function setup_dirs(tmpdir)
         srcdir = joinpath(tmpdir, "src")
         hidden_srcdir = joinpath(tmpdir, ".hidden_srcdir")
@@ -508,8 +450,8 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         return test_src_paths, test_cp_paths
     end
 
-    function cp_follow_symlinks_false_check(s, d; remove_destination=false)
-        cp(s, d; remove_destination=remove_destination, follow_symlinks=false)
+    function cp_follow_symlinks_false_check(s, d; force=false)
+        cp(s, d; force=force, follow_symlinks=false)
         @test isdir(s) == isdir(d)
         @test islink(s) == islink(d)
         islink(s) && @test readlink(s) == readlink(d)
@@ -519,12 +461,12 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         @test length(readdir(d)) == 1
     end
 
-    function mv_check(s, d, d_mv; remove_destination=true)
+    function mv_check(s, d, d_mv; force=true)
         # setup dest
-        cp(s, d; remove_destination=true, follow_symlinks=false)
+        cp(s, d; force=true, follow_symlinks=false)
         stat_d = stat(d)
         # mv(rename) dst to dst_mv
-        mv(d, d_mv; remove_destination=remove_destination)
+        mv(d, d_mv; force=force)
         stat_d_mv = stat(d_mv)
         # make sure d does not exist anymore
         @test !ispath(d)
@@ -540,7 +482,7 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         @test Base.samefile(stat_d, stat_d_mv)
     end
 
-    ## Test require `remove_destination=true` (remove destination first) for existing
+    ## Test require `force=true` (remove destination first) for existing
     #  directories and existing links to directories
     # cp ----------------------------------------------------
     mktempdir() do tmpdir
@@ -557,24 +499,24 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         for (s, d) in zip(test_src_paths2, test_new_paths2)
             cp_follow_symlinks_false_check(s, d)
         end
-        # Test require `remove_destination=true`
+        # Test require `force=true`
         for s in test_src_paths1
             for d in test_new_paths2
-                @test_throws ArgumentError cp(s, d; remove_destination=false)
-                @test_throws ArgumentError cp(s, d; remove_destination=false, follow_symlinks=true)
+                @test_throws ArgumentError cp(s, d; force=false)
+                @test_throws ArgumentError cp(s, d; force=false, follow_symlinks=true)
             end
         end
         # Test remove the existing path first and copy
         # need to use here the test_src_paths2:
         # otherwise ArgumentError: 'src' and 'dst' refer to the same file/dir.
         for (s, d) in zip(test_src_paths2, test_new_paths1)
-            cp_follow_symlinks_false_check(s, d; remove_destination=true)
+            cp_follow_symlinks_false_check(s, d; force=true)
         end
         # Test remove the existing path first and copy an empty dir
         emptydir = joinpath(maindir1, "emptydir")
         mkdir(emptydir)
         for d in test_new_paths1
-            cp(emptydir, d; remove_destination=true, follow_symlinks=false)
+            cp(emptydir, d; force=true, follow_symlinks=false)
             # Expect no link because a dir is copied (follow_symlinks=false does not effect this)
             @test isdir(d) && !islink(d)
             # none should contain any file
@@ -591,16 +533,16 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         test_src_paths1, test_new_paths1 = setup_dirs(maindir1)
         test_src_paths2, test_new_paths2 = setup_dirs(maindir2)
         for (s, d) in zip(test_src_paths1, test_new_paths1)
-            cp_follow_symlinks_false_check(s, d; remove_destination=true)
+            cp_follow_symlinks_false_check(s, d; force=true)
         end
         for (s, d) in zip(test_src_paths2, test_new_paths2)
-            cp_follow_symlinks_false_check(s, d; remove_destination=true)
+            cp_follow_symlinks_false_check(s, d; force=true)
         end
 
-        # Test require `remove_destination=true`
+        # Test require `force=true`
         for s in test_src_paths1
             for d in test_new_paths2
-                @test_throws ArgumentError mv(s, d; remove_destination=false)
+                @test_throws ArgumentError mv(s, d; force=false)
             end
         end
         # Test remove the existing path first and move
@@ -608,7 +550,7 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         # otherwise ArgumentError: 'src' and 'dst' refer to the same file/dir.This is not supported.
         for (s, d) in zip(test_src_paths2, test_new_paths1)
             d_mv = joinpath(dirname(d), "$(basename(d))_mv")
-            mv_check(s, d, d_mv; remove_destination=true)
+            mv_check(s, d, d_mv; force=true)
         end
     end
 
@@ -645,7 +587,7 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         # move the 3 maindirs
         for d in [maindir_new, maindir_new_keepsym, maindir]
             d_mv = joinpath(dirname(d), "$(basename(d))_mv")
-            mv(d, d_mv; remove_destination=true)
+            mv(d, d_mv; force=true)
         end
     end
 
@@ -658,13 +600,13 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         for src in dirs
             for dst in dirs
                 # cptree
-                @test_throws ArgumentError Base.cptree(src,dst; remove_destination=true, follow_symlinks=false)
-                @test_throws ArgumentError Base.cptree(src,dst; remove_destination=true, follow_symlinks=true)
+                @test_throws ArgumentError Base.cptree(src,dst; force=true, follow_symlinks=false)
+                @test_throws ArgumentError Base.cptree(src,dst; force=true, follow_symlinks=true)
                 # cp
-                @test_throws ArgumentError cp(src,dst; remove_destination=true, follow_symlinks=false)
-                @test_throws ArgumentError cp(src,dst; remove_destination=true, follow_symlinks=true)
+                @test_throws ArgumentError cp(src,dst; force=true, follow_symlinks=false)
+                @test_throws ArgumentError cp(src,dst; force=true, follow_symlinks=true)
                 # mv
-                @test_throws ArgumentError mv(src,dst; remove_destination=true)
+                @test_throws ArgumentError mv(src,dst; force=true)
             end
         end
     end
@@ -674,19 +616,19 @@ if !is_windows() || Sys.windows_version() >= Sys.WINDOWS_VISTA_VER
         dst = joinpath(tmpdir, "dst")
         @test !ispath(none_existing_src)
         # cptree
-        @test_throws ArgumentError Base.cptree(none_existing_src,dst; remove_destination=true, follow_symlinks=false)
-        @test_throws ArgumentError Base.cptree(none_existing_src,dst; remove_destination=true, follow_symlinks=true)
+        @test_throws ArgumentError Base.cptree(none_existing_src,dst; force=true, follow_symlinks=false)
+        @test_throws ArgumentError Base.cptree(none_existing_src,dst; force=true, follow_symlinks=true)
         # cp
-        @test_throws Base.UVError cp(none_existing_src,dst; remove_destination=true, follow_symlinks=false)
-        @test_throws Base.UVError cp(none_existing_src,dst; remove_destination=true, follow_symlinks=true)
+        @test_throws Base.IOError cp(none_existing_src,dst; force=true, follow_symlinks=false)
+        @test_throws Base.IOError cp(none_existing_src,dst; force=true, follow_symlinks=true)
         # mv
-        @test_throws Base.UVError mv(none_existing_src,dst; remove_destination=true)
+        @test_throws Base.IOError mv(none_existing_src,dst; force=true)
     end
 end
 
 # issue #10506 #10434
 ## Tests for files and links to files as well as directories and links to directories
-if !is_windows()
+if !Sys.iswindows()
     function setup_files(tmpdir)
         srcfile = joinpath(tmpdir, "srcfile.txt")
         hidden_srcfile = joinpath(tmpdir, ".hidden_srcfile.txt")
@@ -711,22 +653,22 @@ if !is_windows()
         return test_src_paths, test_new_paths, file_txt
     end
 
-    function cp_follow_symlinks_false_check(s, d, file_txt; remove_destination=false)
-        cp(s, d; remove_destination=remove_destination, follow_symlinks=false)
+    function cp_follow_symlinks_false_check(s, d, file_txt; force=false)
+        cp(s, d; force=force, follow_symlinks=false)
         @test isfile(s) == isfile(d)
         @test islink(s) == islink(d)
         islink(s) && @test readlink(s) == readlink(d)
         islink(s) && @test isabspath(readlink(s)) == isabspath(readlink(d))
         # all should contain the same
-        @test readstring(s) == readstring(d) == file_txt
+        @test read(s, String) == read(d, String) == file_txt
     end
 
-    function mv_check(s, d, d_mv, file_txt; remove_destination=true)
+    function mv_check(s, d, d_mv, file_txt; force=true)
         # setup dest
-        cp(s, d; remove_destination=true, follow_symlinks=false)
+        cp(s, d; force=true, follow_symlinks=false)
         stat_d = stat(d)
         # mv(rename) dst to dst_mv
-        mv(d, d_mv; remove_destination=remove_destination)
+        mv(d, d_mv; force=force)
         stat_d_mv = stat(d_mv)
         # make sure d does not exist anymore
         @test !ispath(d)
@@ -736,12 +678,12 @@ if !is_windows()
         islink(s) && @test readlink(s) == readlink(d_mv)
         islink(s) && @test isabspath(readlink(s)) == isabspath(readlink(d_mv))
         # all should contain the same
-        @test readstring(s) == readstring(d_mv) == file_txt
+        @test read(s, String) == read(d_mv, String) == file_txt
         # d => d_mv same file/dir
         @test Base.samefile(stat_d, stat_d_mv)
     end
 
-    ## Test require `remove_destination=true` (remove destination first) for existing
+    ## Test require `force=true` (remove destination first) for existing
     #  files and existing links to files
     # cp ----------------------------------------------------
     mktempdir() do tmpdir
@@ -758,29 +700,29 @@ if !is_windows()
         for (s, d) in zip(test_src_paths2, test_new_paths2)
             cp_follow_symlinks_false_check(s, d, file_txt2)
         end
-        # Test require `remove_destination=true`
+        # Test require `force=true`
         for s in test_src_paths1
             for d in test_new_paths2
-                @test_throws ArgumentError cp(s, d; remove_destination=false)
-                @test_throws ArgumentError cp(s, d; remove_destination=false, follow_symlinks=true)
+                @test_throws ArgumentError cp(s, d; force=false)
+                @test_throws ArgumentError cp(s, d; force=false, follow_symlinks=true)
             end
         end
         # Test remove the existing path first and copy
         # need to use here the test_src_paths2:
         # otherwise ArgumentError: 'src' and 'dst' refer to the same file/dir.This is not supported.
         for (s, d) in zip(test_src_paths2, test_new_paths1)
-            cp_follow_symlinks_false_check(s, d, file_txt2; remove_destination=true)
+            cp_follow_symlinks_false_check(s, d, file_txt2; force=true)
         end
         # Test remove the existing path first and copy an other file
         otherfile = joinpath(tmpdir, "otherfile.txt")
         otherfile_content = "This is otherfile.txt with unicode - 这是一个文件"
         write(otherfile, otherfile_content)
         for d in test_new_paths1
-            cp(otherfile, d; remove_destination=true, follow_symlinks=false)
+            cp(otherfile, d; force=true, follow_symlinks=false)
             # Expect no link because a file is copied (follow_symlinks=false does not effect this)
             @test isfile(d) && !islink(d)
             # all should contain otherfile_content
-            @test readstring(d) == otherfile_content
+            @test read(d, String) == otherfile_content
         end
     end
     # mv ----------------------------------------------------
@@ -798,10 +740,10 @@ if !is_windows()
         for (s, d) in zip(test_src_paths2, test_new_paths2)
             cp_follow_symlinks_false_check(s, d, file_txt2)
         end
-        # Test require `remove_destination=true`
+        # Test require `force=true`
         for s in test_src_paths1
             for d in test_new_paths2
-                @test_throws ArgumentError mv(s, d; remove_destination=false)
+                @test_throws ArgumentError mv(s, d; force=false)
             end
         end
         # Test remove the existing path first and move
@@ -809,7 +751,7 @@ if !is_windows()
         # otherwise ArgumentError: 'src' and 'dst' refer to the same file/dir.This is not supported.
         for (s, d) in zip(test_src_paths2, test_new_paths1)
             d_mv = joinpath(dirname(d), "$(basename(d))_mv")
-            mv_check(s, d, d_mv, file_txt2; remove_destination=true)
+            mv_check(s, d, d_mv, file_txt2; force=true)
         end
     end
 
@@ -840,7 +782,7 @@ if !is_windows()
         rel_dl = "rel_linkto_targetdir"
         rel_dir = joinpath("..", "targetdir")
         symlink(rel_dir, rel_dl)
-        rel_file_read_txt = readstring(rel_file)
+        rel_file_read_txt = read(rel_file, String)
         cd(pwd_)
         # Setup copytodir
         copytodir = joinpath(tmpdir, "copytodir")
@@ -872,7 +814,7 @@ if !is_windows()
         # As expected this will leave some absolute links broken #11145#issuecomment-99315168
         for d in [copytodir, maindir_new, maindir_new_keepsym, maindir]
             d_mv = joinpath(dirname(d), "$(basename(d))_mv")
-            mv(d, d_mv; remove_destination=true)
+            mv(d, d_mv; force=true)
         end
     end
     # issue  ----------------------------------------------------
@@ -884,13 +826,13 @@ if !is_windows()
         for src in files
             for dst in files
                 # cptree
-                @test_throws ArgumentError Base.cptree(src,dst; remove_destination=true, follow_symlinks=false)
-                @test_throws ArgumentError Base.cptree(src,dst; remove_destination=true, follow_symlinks=true)
+                @test_throws ArgumentError Base.cptree(src,dst; force=true, follow_symlinks=false)
+                @test_throws ArgumentError Base.cptree(src,dst; force=true, follow_symlinks=true)
                 # cp
-                @test_throws ArgumentError cp(src,dst; remove_destination=true, follow_symlinks=false)
-                @test_throws ArgumentError cp(src,dst; remove_destination=true, follow_symlinks=true)
+                @test_throws ArgumentError cp(src,dst; force=true, follow_symlinks=false)
+                @test_throws ArgumentError cp(src,dst; force=true, follow_symlinks=true)
                 # mv
-                @test_throws ArgumentError mv(src,dst; remove_destination=true)
+                @test_throws ArgumentError mv(src,dst; force=true)
             end
         end
     end
@@ -903,8 +845,8 @@ end
 ###################
 
 function test_LibcFILE(FILEp)
-    buf = Array{UInt8}(8)
-    str = ccall(:fread, Csize_t, (Ptr{Void}, Csize_t, Csize_t, Ptr{Void}), buf, 1, 8, FILEp)
+    buf = Vector{UInt8}(undef, 8)
+    str = ccall(:fread, Csize_t, (Ptr{Cvoid}, Csize_t, Csize_t, Ptr{Cvoid}), buf, 1, 8, FILEp)
     @test String(buf) == "Hello, w"
     @test position(FILEp) == 8
     seek(FILEp, 5)
@@ -916,9 +858,9 @@ let f = open(file, "w")
     write(f, "Hello, world!")
     close(f)
     f = open(file, "r")
-    test_LibcFILE(convert(Libc.FILE, f))
+    test_LibcFILE(Libc.FILE(f))
     close(f)
-    if is_windows()
+    if Sys.iswindows()
         f = RawFD(ccall(:_open, Cint, (Cstring, Cint), file, Base.Filesystem.JL_O_RDONLY))
     else
         f = RawFD(ccall(:open, Cint, (Cstring, Cint), file, Base.Filesystem.JL_O_RDONLY))
@@ -930,14 +872,15 @@ end
 for f in (mkdir, cd, Base.Filesystem.unlink, readlink, rm, touch, readdir, mkpath,
         stat, lstat, ctime, mtime, filemode, filesize, uperm, gperm, operm, touch,
         isblockdev, ischardev, isdir, isfifo, isfile, islink, ispath, issetgid,
-        issetuid, issocket, issticky, realpath, watch_file, poll_file)
+        issetuid, issocket, issticky, realpath)
+    local f
     @test_throws ArgumentError f("adir\0bad")
 end
 @test_throws ArgumentError chmod("ba\0d", 0o222)
 @test_throws ArgumentError open("ba\0d", "w")
 @test_throws ArgumentError cp(file, "ba\0d")
 @test_throws ArgumentError mv(file, "ba\0d")
-if !is_windows() || (Sys.windows_version() >= Sys.WINDOWS_VISTA_VER)
+if !Sys.iswindows() || (Sys.windows_version() >= Sys.WINDOWS_VISTA_VER)
     @test_throws ArgumentError symlink(file, "ba\0d")
 else
     @test_throws ErrorException symlink(file, "ba\0d")
@@ -959,7 +902,7 @@ cd(dirwalk) do
         touch(joinpath("sub_dir1", "file$i"))
     end
     touch(joinpath("sub_dir2", "file_dir2"))
-    has_symlinks = !is_windows() || (Sys.windows_version() >= Sys.WINDOWS_VISTA_VER)
+    has_symlinks = !Sys.iswindows() || (Sys.windows_version() >= Sys.WINDOWS_VISTA_VER)
     follow_symlink_vec = has_symlinks ? [true, false] : [false]
     has_symlinks && symlink(abspath("sub_dir2"), joinpath("sub_dir1", "link"))
     for follow_symlinks in follow_symlink_vec
@@ -1048,11 +991,11 @@ rm(dirwalk, recursive=true)
 ############
 # Clean up #
 ############
-if !is_windows()
+if !Sys.iswindows()
     rm(link)
     rm(rellink)
 end
-if !is_windows() || (Sys.windows_version() >= Sys.WINDOWS_VISTA_VER)
+if !Sys.iswindows() || (Sys.windows_version() >= Sys.WINDOWS_VISTA_VER)
     rm(dirlink)
     rm(relsubdirlink)
 end
@@ -1064,6 +1007,54 @@ rm(dir)
 @test !ispath(file)
 @test !ispath(dir)
 
+
+
+##################
+# Return values of mkpath, mkdir, cp, mv and touch
+####################
+mktempdir() do dir
+    name1 = joinpath(dir, "apples")
+    name2 = joinpath(dir, "bannanas")
+    @test !ispath(name1)
+    @test touch(name1) == name1
+    @test isfile(name1)
+    @test touch(name1) == name1
+    @test isfile(name1)
+    @test !ispath(name2)
+    @test mv(name1, name2) == name2
+    @test !ispath(name1)
+    @test isfile(name2)
+    @test cp(name2, name1) == name1
+    @test isfile(name1)
+    @test isfile(name2)
+    namedir = joinpath(dir, "chalk")
+    namepath = joinpath(dir, "chalk","cheese","fresh")
+    @test !ispath(namedir)
+    @test mkdir(namedir) == namedir
+    @test isdir(namedir)
+    @test !ispath(namepath)
+    @test mkpath(namepath) == namepath
+    @test isdir(namepath)
+    @test mkpath(namepath) == namepath
+    @test isdir(namepath)
+end
+
+# issue #30588
+@test realpath(".") == realpath(pwd())
+mktempdir() do dir
+    cd(dir) do
+        path = touch("FooBar.txt")
+        @test ispath(realpath(path))
+        if ispath(uppercase(path)) # case-insensitive filesystem
+            @test realpath(path) == realpath(uppercase(path)) == realpath(lowercase(path)) ==
+                  realpath(uppercase(realpath(path))) == realpath(lowercase(realpath(path)))
+            @test basename(realpath(uppercase(path))) == path
+        end
+        rm(path)
+        @test_throws SystemError realpath(path)
+    end
+end
+
 # issue #9687
 let n = tempname()
     w = open(n, "a")
@@ -1077,60 +1068,50 @@ let n = tempname()
     rm(n)
 end
 
-#issue #12992
-function test_12992()
-    pfw = PollingFileWatcher(@__FILE__, 0.01)
-    close(pfw)
-    pfw = PollingFileWatcher(@__FILE__, 0.01)
-    close(pfw)
-    pfw = PollingFileWatcher(@__FILE__, 0.01)
-    close(pfw)
-    gc()
-    gc()
-end
+@test_throws ArgumentError mkpath("fakepath", mode = -1)
 
-# Make sure multiple close is fine
-function test2_12992()
-    pfw = PollingFileWatcher(@__FILE__, 0.01)
-    close(pfw)
-    close(pfw)
-    pfw = PollingFileWatcher(@__FILE__, 0.01)
-    close(pfw)
-    close(pfw)
-    pfw = PollingFileWatcher(@__FILE__, 0.01)
-    close(pfw)
-    close(pfw)
-    gc()
-    gc()
-end
+@testset "mktempdir 'prefix' argument" begin
+    tmpdirbase = joinpath(tempdir(), "")
+    def_prefix = "jl_"
+    mktempdir() do tmpdir
+        @test isdir(tmpdir)
+        @test startswith(tmpdir, tmpdirbase * def_prefix)
+        @test sizeof(tmpdir) == sizeof(tmpdirbase) + sizeof(def_prefix) + 6
+        @test sizeof(basename(tmpdir)) == sizeof(def_prefix) + 6
+        cd(tmpdir) do
+            Sys.iswindows() || mkdir(".\\")
+            for relpath in (".", "./", ".\\", "")
+                mktempdir(relpath) do tmpdir2
+                    pfx = joinpath(relpath, def_prefix)
+                    @test sizeof(tmpdir2) == sizeof(pfx) + 6
+                    @test startswith(tmpdir2, pfx)
+                end
+            end
+        end
+    end
+    # Special character prefix tests
+    for tst_prefix in ("ABCDEF", "./pfx", ".\\pfx", "", "#!@%^&()-", "/", "\\", "////abc", "\\\\\\\\abc", "∃x∀y")
+        mktempdir(; prefix=tst_prefix) do tmpdir
+            @test isdir(tmpdir)
+            @test startswith(tmpdir, tmpdirbase * tst_prefix)
+            @test sizeof(basename(tmpdir)) == 6 + sizeof(basename(tst_prefix))
+        end
+    end
 
-test_12992()
-test_12992()
-test_12992()
+    @test_throws Base.IOError mktempdir(; prefix="dir_notexisting/bar")
+    @test_throws Base.IOError mktempdir(; prefix="dir_notexisting/")
+    @test_throws Base.IOError mktempdir("dir_notexisting/")
 
-test2_12992()
-test2_12992()
-test2_12992()
+    # Behavioral differences across OS types
+    if Sys.iswindows()
+        # invalid file name
+        @test_throws Base.IOError mktempdir(; prefix="a*b")
+        @test_throws Base.IOError mktempdir("a*b")
+    end
 
-# issue 13559
-if !is_windows()
-function test_13559()
-    fn = tempname()
-    run(`mkfifo $fn`)
-    # use subprocess to write 127 bytes to FIFO
-    writer_cmds = "x=open(\"$fn\", \"w\"); for i=1:127 write(x,0xaa); flush(x); sleep(0.1) end; close(x); quit()"
-    open(pipeline(`$(Base.julia_cmd()) --startup-file=no -e $writer_cmds`, stderr=STDERR))
-    #quickly read FIFO, draining it and blocking but not failing with EOFError yet
-    r = open(fn, "r")
-    # 15 proper reads
-    for i=1:15
-        @test read(r, Int64) == -6148914691236517206
+    mktempdir(""; prefix=tmpdirbase) do tmpdir
+        @test startswith(tmpdir, tmpdirbase)
+        @test sizeof(tmpdir) == 6 + sizeof(tmpdirbase)
+        @test sizeof(basename(tmpdir)) == 6
     end
-    # last read should throw EOFError when FIFO closes, since there are only 7 bytes available.
-    @test_throws EOFError read(r, Int64)
-    close(r)
-    rm(fn)
 end
-test_13559()
-end
-@test_throws ArgumentError mkpath("fakepath",-1)
